import logging
from collections.abc import Generator
from typing import ClassVar, Final

from flask import Response

from .input_fields import FieldId

logger: logging.Logger | None

HTML_HEAD: Final[str]
HTML_TAIL: Final[str]

class Executor:
    RAW_CMD_PATH: ClassVar[str]
    returncode: int | None

    def __init__(self) -> None: ...
    def exec(self, command_path: str) -> Response: ...
    def _exec_raw(self, command: list[str]) -> Response: ...  # undocumented
    def _exec_html(self, command_path: str) -> Response: ...  # undocumented
    def _run_script_and_generate_stream(self) -> Generator[str]: ...  # undocumented
    def _create_cmd_header(self, commands: list[CmdPart]) -> str: ...  # undocumented
    def _create_result_footer(self) -> Generator[str]: ...  # undocumented

def _get_download_link(field_info: FieldFileInfo) -> str: ...  # undocumented

class CommandLineRaw:
    def __init__(self, script_file_path: str, command: str) -> None: ...
    def append(self, part: str, secret: bool = False) -> None: ...
    def get_commandline(self, obfuscate: bool = False) -> list[str]: ...
    def get_download_field_infos(self) -> list[FieldInfo]: ...
    def after_script_executed(self) -> None: ...

class CommandLineForm:
    command_line_bulder: FormToCommandLineBuilder
    def __init__(self, script_file_path: str, commands: list[str]) -> None: ...
    def append(self, part: str, secret: bool = False) -> None: ...
    def get_commandline(self, obfuscate: bool = False) -> list[str]: ...
    def get_download_field_infos(self) -> list[FieldInfo]: ...
    def after_script_executed(self) -> None: ...

def _get_python_interpreter() -> str: ...

class CmdPart:
    def __init__(self, part: str, secret: bool = False) -> None: ...

class FormToCommandLineBuilder:
    def __init__(self, command_line: CommandLineForm) -> None: ...
    def add_command_args(self, command_index: int) -> None: ...
    @staticmethod
    def _is_option(cmd_option: str) -> bool: ...
    def _process_option(self, field_info: FieldInfo) -> None: ...

class FieldInfo:
    param: FieldId
    key: str
    is_file: bool
    cmd_opt: str
    generate_download_link: bool
    @staticmethod
    def factory(key: str) -> FieldInfo: ...
    def __init__(self, param: FieldId) -> None: ...
    def before_script_execute(self) -> None: ...
    def after_script_executed(self) -> None: ...
    def __lt__(self, other: object) -> bool: ...
    def __eq__(self, other: object) -> bool: ...

class FieldFileInfo(FieldInfo):
    mode: str
    generate_download_link: bool
    link_name: str
    file_path: str
    def __init__(self, fimeta: FieldId) -> None: ...
    def before_script_execute(self) -> None: ...
    @classmethod
    def temp_dir(cls) -> str: ...
    def save(self) -> None: ...

class FieldOutFileInfo(FieldFileInfo):
    file_suffix: str
    def __init__(self, fimeta: FieldId) -> None: ...
    def save(self) -> None: ...

class FieldPathInfo(FieldFileInfo):
    def save(self) -> None: ...
    def after_script_executed(self) -> None: ...

class FieldPathOutInfo(FieldOutFileInfo):
    def save(self) -> None: ...
    def after_script_executed(self) -> None: ...

def zip_folder(folder_path: str, out_folder: str, out_prefix: str) -> str: ...



from _typeshed import Incomplete
from typing import ClassVar
from typing_extensions import Self

import passlib.utils.handlers as uh

class scram(uh.HasRounds, uh.HasRawSalt, uh.HasRawChecksum, uh.GenericHandler):  # type: ignore[misc]
    name: ClassVar[str]
    ident: ClassVar[str]
    default_salt_size: ClassVar[int]
    max_salt_size: ClassVar[int]
    default_rounds: ClassVar[int]
    min_rounds: ClassVar[int]
    max_rounds: ClassVar[int]
    rounds_cost: ClassVar[str]
    default_algs: ClassVar[list[str]]
    algs: Incomplete | None
    @classmethod
    def extract_digest_info(cls, hash, alg): ...
    @classmethod
    def extract_digest_algs(cls, hash, format: str = "iana"): ...
    @classmethod
    def derive_digest(cls, password, salt, rounds, alg): ...
    @classmethod
    def from_string(cls, hash) -> Self: ...  # type: ignore[override]
    @classmethod
    def using(cls, default_algs=None, algs=None, **kwds): ...  # type: ignore[override]
    def __init__(self, algs=None, **kwds) -> None: ...
    @classmethod
    def verify(cls, secret, hash, full: bool = False): ...  # type: ignore[override]

__all__ = ["scram"]



from _typeshed import Incomplete

class JsonWebSignature:
    REGISTERED_HEADER_PARAMETER_NAMES: frozenset[str]
    MAX_CONTENT_LENGTH: int
    ALGORITHMS_REGISTRY: dict[str, Incomplete]
    def __init__(self, algorithms=None, private_headers=None) -> None: ...
    @classmethod
    def register_algorithm(cls, algorithm) -> None: ...
    def serialize_compact(self, protected, payload, key): ...
    def deserialize_compact(self, s, key, decode=None): ...
    def serialize_json(self, header_obj, payload, key): ...
    def deserialize_json(self, obj, key, decode=None): ...
    def serialize(self, header, payload, key): ...
    def deserialize(self, s, key, decode=None): ...



from _typeshed import Incomplete
from typing import Literal, overload
from typing_extensions import TypeAlias

import numpy as np
from numpy.typing import NDArray

from ._enum import ParamEnum
from ._ragged_array import from_ragged_array as from_ragged_array, to_ragged_array as to_ragged_array
from ._typing import ArrayLikeSeq, GeoArray, OptGeoArrayLikeSeq
from .geometry.base import BaseGeometry
from .lib import Geometry

__all__ = ["from_geojson", "from_ragged_array", "from_wkb", "from_wkt", "to_geojson", "to_ragged_array", "to_wkb", "to_wkt"]

_OutputDimension: TypeAlias = Literal[2, 3, 4]

# Mypy and stubtest aren't happy with the following definition and
# raise is a reserved keyword, so we cannot use the class syntax of enums
# DecodingErrorOptions = ParamEnum("DecodingErrorOptions", {"ignore": 0, "warn": 1, "raise": 2, "fix": 3})
DecodingErrorOptions: Incomplete

class WKBFlavorOptions(ParamEnum):
    extended = 1
    iso = 2

@overload
def to_wkt(
    geometry: None,
    rounding_precision: int = 6,
    trim: bool = True,
    output_dimension: _OutputDimension | None = None,
    old_3d: bool = False,
    **kwargs,
) -> None: ...
@overload
def to_wkt(
    geometry: Geometry,
    rounding_precision: int = 6,
    trim: bool = True,
    output_dimension: _OutputDimension | None = None,
    old_3d: bool = False,
    **kwargs,
) -> str: ...
@overload
def to_wkt(
    geometry: OptGeoArrayLikeSeq,
    rounding_precision: int = 6,
    trim: bool = True,
    output_dimension: _OutputDimension | None = None,
    old_3d: bool = False,
    **kwargs,
) -> NDArray[np.str_]: ...
@overload
def to_wkb(
    geometry: None,
    hex: bool = False,
    output_dimension: _OutputDimension | None = None,
    byte_order: int = -1,
    include_srid: bool = False,
    flavor: Literal["iso", "extended"] = "extended",
    **kwargs,
) -> None: ...
@overload
def to_wkb(
    geometry: Geometry,
    hex: Literal[False] = False,
    output_dimension: _OutputDimension | None = None,
    byte_order: int = -1,
    include_srid: bool = False,
    flavor: Literal["iso", "extended"] = "extended",
    **kwargs,
) -> bytes: ...
@overload
def to_wkb(
    geometry: Geometry,
    hex: Literal[True],
    output_dimension: _OutputDimension | None = None,
    byte_order: int = -1,
    include_srid: bool = False,
    flavor: Literal["iso", "extended"] = "extended",
    **kwargs,
) -> str: ...
@overload
def to_wkb(
    geometry: Geometry,
    hex: bool,
    output_dimension: _OutputDimension | None = None,
    byte_order: int = -1,
    include_srid: bool = False,
    flavor: Literal["iso", "extended"] = "extended",
    **kwargs,
) -> bytes | str: ...
@overload
def to_wkb(
    geometry: OptGeoArrayLikeSeq,
    hex: Literal[False] = False,
    output_dimension: _OutputDimension | None = None,
    byte_order: int = -1,
    include_srid: bool = False,
    flavor: Literal["iso", "extended"] = "extended",
    **kwargs,
) -> NDArray[np.bytes_]: ...
@overload
def to_wkb(
    geometry: OptGeoArrayLikeSeq,
    hex: Literal[True],
    output_dimension: _OutputDimension | None = None,
    byte_order: int = -1,
    include_srid: bool = False,
    flavor: Literal["iso", "extended"] = "extended",
    **kwargs,
) -> NDArray[np.str_]: ...
@overload
def to_wkb(
    geometry: OptGeoArrayLikeSeq,
    hex: bool,
    output_dimension: _OutputDimension | None = None,
    byte_order: int = -1,
    include_srid: bool = False,
    flavor: Literal["iso", "extended"] = "extended",
    **kwargs,
) -> NDArray[np.bytes_] | NDArray[np.str_]: ...
@overload
def to_geojson(geometry: None, indent: int | None = None, **kwargs) -> None: ...
@overload
def to_geojson(geometry: Geometry, indent: int | None = None, **kwargs) -> str: ...
@overload
def to_geojson(geometry: OptGeoArrayLikeSeq, indent: int | None = None, **kwargs) -> NDArray[np.str_]: ...
@overload
def from_wkt(geometry: None, on_invalid: Literal["raise", "warn", "ignore", "fix"] = "raise", **kwargs) -> None: ...
@overload
def from_wkt(geometry: str, on_invalid: Literal["raise", "warn", "ignore", "fix"] = "raise", **kwargs) -> BaseGeometry: ...  # type: ignore[overload-overlap]
@overload
def from_wkt(
    geometry: ArrayLikeSeq[str | None], on_invalid: Literal["raise", "warn", "ignore", "fix"] = "raise", **kwargs
) -> GeoArray: ...
@overload
def from_wkb(geometry: None, on_invalid: Literal["raise", "warn", "ignore", "fix"] = "raise", **kwargs) -> None: ...
@overload
def from_wkb(
    geometry: str | bytes, on_invalid: Literal["raise", "warn", "ignore", "fix"] = "raise", **kwargs
) -> BaseGeometry: ...  # type: ignore[overload-overlap]
@overload
def from_wkb(
    geometry: ArrayLikeSeq[str | bytes | None], on_invalid: Literal["raise", "warn", "ignore", "fix"] = "raise", **kwargs
) -> GeoArray: ...
@overload
def from_geojson(geometry: None, on_invalid: Literal["raise", "warn", "ignore", "fix"] = "raise", **kwargs) -> None: ...
@overload
def from_geojson(
    geometry: str | bytes, on_invalid: Literal["raise", "warn", "ignore", "fix"] = "raise", **kwargs
) -> BaseGeometry: ...  # type: ignore[overload-overlap]
@overload
def from_geojson(
    geometry: ArrayLikeSeq[str | bytes | None], on_invalid: Literal["raise", "warn", "ignore", "fix"] = "raise", **kwargs
) -> GeoArray: ...



from typing import Final

DEBUG: Final[bool]
STATIC_DIR: Final[None]
CUT_TRACEBACK: Final[bool]
STD_DOCTYPE: Final[str]
STD_STYLESHEETS: Final[list[tuple[str, ...]]]
BASE_STYLESHEETS_PLACEHOLDER: Final[str]
COMPONENT_STYLESHEETS_PLACEHOLDER: Final[str]
SCRIPTS_PLACEHOLDER: Final[str]
RELOADING_CHECK_INTERVAL: Final[float]
LOG_TO_SQLITE: Final[None]
LOGGING_LEVEL: Final[None]
LOGGING_PONY_LEVEL: Final[None]
MAX_SESSION_CTIME: Final[int]
MAX_SESSION_MTIME: Final[int]
MAX_LONGLIFE_SESSION: Final[int]
COOKIE_SERIALIZATION_TYPE: Final[str]
COOKIE_NAME: Final[str]
COOKIE_PATH: Final[str]
COOKIE_DOMAIN: Final[None]
HASH_ALGORITHM: Final[None]
SESSION_STORAGE: Final[None]
MEMCACHE: Final[None]
ALTERNATIVE_SESSION_MEMCACHE: Final[None]
ALTERNATIVE_ORM_MEMCACHE: Final[None]
ALTERNATIVE_TEMPLATING_MEMCACHE: Final[None]
ALTERNATIVE_RESPONSE_MEMCACHE: Final[None]
PICKLE_START_OFFSET: Final[int]
PICKLE_HTML_AS_PLAIN_STR: Final[bool]
RESTORE_ESCAPES: Final[bool]
SOURCE_ENCODING: Final[None]
CONSOLE_ENCODING: Final[None]
MAX_FETCH_COUNT: Final[None]
CONSOLE_WIDTH: Final[int]
SIMPLE_ALIASES: Final[bool]
INNER_JOIN_SYNTAX: Final[bool]
DEBUGGING_REMOVE_ADDR: Final[bool]
DEBUGGING_RESTORE_ESCAPES: Final[bool]



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

import builtins
import collections.abc
import typing

import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class MemAllocatorStats(google.protobuf.message.Message):
    """Some of the data from AllocatorStats"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NUM_ALLOCS_FIELD_NUMBER: builtins.int
    BYTES_IN_USE_FIELD_NUMBER: builtins.int
    PEAK_BYTES_IN_USE_FIELD_NUMBER: builtins.int
    LARGEST_ALLOC_SIZE_FIELD_NUMBER: builtins.int
    FRAGMENTATION_METRIC_FIELD_NUMBER: builtins.int
    num_allocs: builtins.int
    bytes_in_use: builtins.int
    peak_bytes_in_use: builtins.int
    largest_alloc_size: builtins.int
    fragmentation_metric: builtins.float
    def __init__(
        self,
        *,
        num_allocs: builtins.int | None = ...,
        bytes_in_use: builtins.int | None = ...,
        peak_bytes_in_use: builtins.int | None = ...,
        largest_alloc_size: builtins.int | None = ...,
        fragmentation_metric: builtins.float | None = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing.Literal[
            "bytes_in_use",
            b"bytes_in_use",
            "fragmentation_metric",
            b"fragmentation_metric",
            "largest_alloc_size",
            b"largest_alloc_size",
            "num_allocs",
            b"num_allocs",
            "peak_bytes_in_use",
            b"peak_bytes_in_use",
        ],
    ) -> None: ...

global___MemAllocatorStats = MemAllocatorStats

@typing.final
class MemChunk(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ADDRESS_FIELD_NUMBER: builtins.int
    SIZE_FIELD_NUMBER: builtins.int
    REQUESTED_SIZE_FIELD_NUMBER: builtins.int
    BIN_FIELD_NUMBER: builtins.int
    OP_NAME_FIELD_NUMBER: builtins.int
    FREED_AT_COUNT_FIELD_NUMBER: builtins.int
    ACTION_COUNT_FIELD_NUMBER: builtins.int
    IN_USE_FIELD_NUMBER: builtins.int
    STEP_ID_FIELD_NUMBER: builtins.int
    address: builtins.int
    size: builtins.int
    requested_size: builtins.int
    bin: builtins.int
    op_name: builtins.str
    freed_at_count: builtins.int
    action_count: builtins.int
    in_use: builtins.bool
    step_id: builtins.int
    def __init__(
        self,
        *,
        address: builtins.int | None = ...,
        size: builtins.int | None = ...,
        requested_size: builtins.int | None = ...,
        bin: builtins.int | None = ...,
        op_name: builtins.str | None = ...,
        freed_at_count: builtins.int | None = ...,
        action_count: builtins.int | None = ...,
        in_use: builtins.bool | None = ...,
        step_id: builtins.int | None = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing.Literal[
            "action_count",
            b"action_count",
            "address",
            b"address",
            "bin",
            b"bin",
            "freed_at_count",
            b"freed_at_count",
            "in_use",
            b"in_use",
            "op_name",
            b"op_name",
            "requested_size",
            b"requested_size",
            "size",
            b"size",
            "step_id",
            b"step_id",
        ],
    ) -> None: ...

global___MemChunk = MemChunk

@typing.final
class BinSummary(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    BIN_FIELD_NUMBER: builtins.int
    TOTAL_BYTES_IN_USE_FIELD_NUMBER: builtins.int
    TOTAL_BYTES_IN_BIN_FIELD_NUMBER: builtins.int
    TOTAL_CHUNKS_IN_USE_FIELD_NUMBER: builtins.int
    TOTAL_CHUNKS_IN_BIN_FIELD_NUMBER: builtins.int
    bin: builtins.int
    total_bytes_in_use: builtins.int
    total_bytes_in_bin: builtins.int
    total_chunks_in_use: builtins.int
    total_chunks_in_bin: builtins.int
    def __init__(
        self,
        *,
        bin: builtins.int | None = ...,
        total_bytes_in_use: builtins.int | None = ...,
        total_bytes_in_bin: builtins.int | None = ...,
        total_chunks_in_use: builtins.int | None = ...,
        total_chunks_in_bin: builtins.int | None = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing.Literal[
            "bin",
            b"bin",
            "total_bytes_in_bin",
            b"total_bytes_in_bin",
            "total_bytes_in_use",
            b"total_bytes_in_use",
            "total_chunks_in_bin",
            b"total_chunks_in_bin",
            "total_chunks_in_use",
            b"total_chunks_in_use",
        ],
    ) -> None: ...

global___BinSummary = BinSummary

@typing.final
class SnapShot(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ACTION_COUNT_FIELD_NUMBER: builtins.int
    SIZE_FIELD_NUMBER: builtins.int
    action_count: builtins.int
    size: builtins.int
    def __init__(self, *, action_count: builtins.int | None = ..., size: builtins.int | None = ...) -> None: ...
    def ClearField(self, field_name: typing.Literal["action_count", b"action_count", "size", b"size"]) -> None: ...

global___SnapShot = SnapShot

@typing.final
class MemoryDump(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ALLOCATOR_NAME_FIELD_NUMBER: builtins.int
    BIN_SUMMARY_FIELD_NUMBER: builtins.int
    CHUNK_FIELD_NUMBER: builtins.int
    SNAP_SHOT_FIELD_NUMBER: builtins.int
    STATS_FIELD_NUMBER: builtins.int
    allocator_name: builtins.str
    @property
    def bin_summary(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___BinSummary]: ...
    @property
    def chunk(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___MemChunk]: ...
    @property
    def snap_shot(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___SnapShot]: ...
    @property
    def stats(self) -> global___MemAllocatorStats: ...
    def __init__(
        self,
        *,
        allocator_name: builtins.str | None = ...,
        bin_summary: collections.abc.Iterable[global___BinSummary] | None = ...,
        chunk: collections.abc.Iterable[global___MemChunk] | None = ...,
        snap_shot: collections.abc.Iterable[global___SnapShot] | None = ...,
        stats: global___MemAllocatorStats | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["stats", b"stats"]) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing.Literal[
            "allocator_name",
            b"allocator_name",
            "bin_summary",
            b"bin_summary",
            "chunk",
            b"chunk",
            "snap_shot",
            b"snap_shot",
            "stats",
            b"stats",
        ],
    ) -> None: ...

global___MemoryDump = MemoryDump



import sys
from signal import _HANDLER, _SIGNUM

# technically the implementations will always be around, but since they always
# throw an exception on windows, due to the missing SIGCHLD, we might as well
# pretent they don't exist, but what is different, is that the parameters are
# named even pre 3.10, so we don't just import the symbol from stdlib signal
if sys.platform != "win32":
    def getsignal(signalnum: _SIGNUM) -> _HANDLER: ...
    def signal(signalnum: _SIGNUM, handler: _HANDLER) -> _HANDLER: ...
    def set_wakeup_fd(fd: int, /, *, warn_on_full_buffer: bool = True) -> int: ...

    __all__ = ["signal", "getsignal", "set_wakeup_fd"]



from __future__ import annotations

from typing import Optional
from typing_extensions import assert_type

import greenlet

g = greenlet.greenlet()
h = greenlet.greenlet()
assert_type(g.parent, Optional[greenlet.greenlet])
g.parent = h

# Although "parent" sometimes can be None at runtime,
# it's always illegal for it to be set to None
g.parent = None  # type: ignore



from collections.abc import Generator

from networkx.classes.digraph import DiGraph
from networkx.classes.graph import _Node
from networkx.utils.backends import _dispatchable

__all__ = ["number_attracting_components", "attracting_components", "is_attracting_component"]

@_dispatchable
def attracting_components(G: DiGraph[_Node]) -> Generator[set[_Node]]: ...
@_dispatchable
def number_attracting_components(G: DiGraph[_Node]) -> int: ...
@_dispatchable
def is_attracting_component(G: DiGraph[_Node]) -> bool: ...



from _typeshed import Incomplete

from networkx.classes.graph import Graph, _Node
from networkx.utils.backends import _dispatchable

__all__ = ["edge_betweenness_partition", "edge_current_flow_betweenness_partition"]

@_dispatchable
def edge_betweenness_partition(G: Graph[_Node], number_of_sets: int, *, weight: str | None = None) -> list[Incomplete]: ...
@_dispatchable
def edge_current_flow_betweenness_partition(
    G: Graph[_Node], number_of_sets: int, *, weight: str | None = None
) -> list[Incomplete]: ...



from typing import Any

class classproperty:
    im_func: Any
    def __init__(self, func) -> None: ...
    def __get__(self, obj, cls): ...
    @property
    def __func__(self): ...

class hybrid_method:
    func: Any
    def __init__(self, func) -> None: ...
    def __get__(self, obj, cls): ...

def memoize_single_value(func): ...

class memoized_property:
    __func__: Any
    __name__: Any
    __doc__: Any
    def __init__(self, func) -> None: ...
    def __get__(self, obj, cls): ...
    def clear_cache(self, obj) -> None: ...
    def peek_cache(self, obj, default=None): ...

def deprecated_function(
    msg=None, deprecated=None, removed=None, updoc: bool = True, replacement=None, _is_method: bool = False, func_module=None
): ...
def deprecated_method(msg=None, deprecated=None, removed=None, updoc: bool = True, replacement=None): ...

__all__ = [
    "classproperty",
    "hybrid_method",
    "memoize_single_value",
    "memoized_property",
    "deprecated_function",
    "deprecated_method",
]



import datetime
from typing import overload

@overload
def list_to_scope(scope: None) -> None: ...
@overload
def list_to_scope(scope: str | set[object] | tuple[object] | list[object]) -> str: ...
@overload
def scope_to_list(scope: None) -> None: ...
@overload
def scope_to_list(scope: str | set[object] | tuple[object] | list[object]) -> list[str]: ...
def params_from_uri(uri: str) -> dict[str, str | list[str]]: ...
def host_from_uri(uri: str) -> tuple[str, str | None]: ...
def escape(u: str) -> str: ...
def generate_age(issue_time: datetime.datetime | datetime.timedelta) -> str: ...
def is_secure_transport(uri: str) -> bool: ...



from collections.abc import Callable, Iterable
from typing import Any

from django import forms
from django.db.models import Q, QuerySet
from django.forms import Field
from django_stubs_ext import StrOrPromise

from .fields import (
    BaseCSVField,
    BaseRangeField,
    DateRangeField,
    DateTimeRangeField,
    IsoDateTimeField,
    IsoDateTimeRangeField,
    Lookup,
    LookupChoiceField,
    ModelChoiceField,
    ModelMultipleChoiceField,
    RangeField,
    TimeRangeField,
)

__all__ = [
    "AllValuesFilter",
    "AllValuesMultipleFilter",
    "BaseCSVFilter",
    "BaseInFilter",
    "BaseRangeFilter",
    "BooleanFilter",
    "CharFilter",
    "ChoiceFilter",
    "DateFilter",
    "DateFromToRangeFilter",
    "DateRangeFilter",
    "DateTimeFilter",
    "DateTimeFromToRangeFilter",
    "DurationFilter",
    "Filter",
    "IsoDateTimeFilter",
    "IsoDateTimeFromToRangeFilter",
    "LookupChoiceFilter",
    "ModelChoiceFilter",
    "ModelMultipleChoiceFilter",
    "MultipleChoiceFilter",
    "NumberFilter",
    "NumericRangeFilter",
    "OrderingFilter",
    "RangeFilter",
    "TimeFilter",
    "TimeRangeFilter",
    "TypedChoiceFilter",
    "TypedMultipleChoiceFilter",
    "UUIDFilter",
]

class Filter:
    creation_counter: int
    field_class: type[Any]  # Subclasses specify more specific field types
    field_name: str | None
    lookup_expr: str
    distinct: bool
    exclude: bool
    extra: dict[str, Any]  # Field kwargs can include various types of parameters
    def __init__(
        self,
        field_name: str | None = None,
        lookup_expr: str | None = None,
        *,
        label: StrOrPromise | None = None,
        method: Callable[..., Any] | str | None = None,  # Filter methods can return various types
        distinct: bool = False,
        exclude: bool = False,
        **kwargs: Any,  # Field kwargs stored as extra (required, help_text, etc.)
    ) -> None: ...
    def get_method(self, qs: QuerySet[Any]) -> Callable[..., QuerySet[Any]]: ...  # Returns QuerySet filtering methods
    method: Callable[..., Any] | str | None  # Custom filter methods return various types
    label: StrOrPromise | None  # Filter label for display
    @property
    def field(self) -> Field: ...
    def filter(self, qs: QuerySet[Any], value: Any) -> QuerySet[Any]: ...  # Filter value can be any user input type

class CharFilter(Filter):
    field_class: type[forms.CharField]

class BooleanFilter(Filter):
    field_class: type[forms.NullBooleanField]

class ChoiceFilter(Filter):
    field_class: type[Any]  # Base class for choice-based filters
    null_value: Any  # Null value can be any type (None, empty string, etc.)
    def __init__(
        self,
        field_name: str | None = None,
        lookup_expr: str | None = None,
        *,
        null_value: Any = ...,  # Null value can be any type (None, empty string, etc.)
        # Inherited from Filter
        label: StrOrPromise | None = None,
        method: Callable[..., Any] | str | None = None,  # Filter methods can return various types
        distinct: bool = False,
        exclude: bool = False,
        **kwargs: Any,  # Field kwargs stored as extra (required, help_text, etc.)
    ) -> None: ...
    def filter(self, qs: QuerySet[Any], value: Any) -> QuerySet[Any]: ...

class TypedChoiceFilter(Filter):
    field_class: type[forms.TypedChoiceField]

class UUIDFilter(Filter):
    field_class: type[forms.UUIDField]

class MultipleChoiceFilter(Filter):
    field_class: type[Any]  # Base class for multiple choice filters
    always_filter: bool
    conjoined: bool
    null_value: Any  # Multiple choice null values vary by implementation
    def __init__(
        self,
        field_name: str | None = None,
        lookup_expr: str | None = None,
        *,
        distinct: bool = True,  # Overrides distinct default
        conjoined: bool = False,
        null_value: Any = ...,  # Multiple choice null values vary by implementation
        # Inherited from Filter
        label: StrOrPromise | None = None,
        method: Callable[..., Any] | str | None = None,  # Filter methods can return various types
        exclude: bool = False,
        **kwargs: Any,  # Field kwargs stored as extra (required, help_text, etc.)
    ) -> None: ...
    def is_noop(self, qs: QuerySet[Any], value: Any) -> bool: ...  # Value can be any filter input
    def filter(self, qs: QuerySet[Any], value: Any) -> QuerySet[Any]: ...
    def get_filter_predicate(self, v: Any) -> Q: ...  # Predicate value can be any filter input type

class TypedMultipleChoiceFilter(MultipleChoiceFilter):
    field_class: type[forms.TypedMultipleChoiceField]  # More specific than parent MultipleChoiceField

class DateFilter(Filter):
    field_class: type[forms.DateField]

class DateTimeFilter(Filter):
    field_class: type[forms.DateTimeField]

class IsoDateTimeFilter(DateTimeFilter):
    field_class: type[IsoDateTimeField]

class TimeFilter(Filter):
    field_class: type[forms.TimeField]

class DurationFilter(Filter):
    field_class: type[forms.DurationField]

class QuerySetRequestMixin:
    queryset: QuerySet[Any] | None
    def __init__(self, *, queryset: QuerySet[Any] | None) -> None: ...
    def get_request(self) -> Any: ...  # Request can be HttpRequest or other request types
    def get_queryset(self, request: Any) -> QuerySet[Any]: ...  # Request parameter accepts various request types
    @property
    def field(self) -> Field: ...

class ModelChoiceFilter(QuerySetRequestMixin, ChoiceFilter):
    field_class: type[ModelChoiceField]  # More specific than parent ChoiceField
    def __init__(
        self,
        field_name: str | None = None,
        lookup_expr: str | None = None,
        *,
        # Inherited from QuerySetRequestMixin
        queryset: QuerySet[Any] | None = None,
        # Inherited from ChoiceFilter
        null_value: Any = ...,  # Null value can be any type (None, empty string, etc.)
        # Inherited from Filter
        label: StrOrPromise | None = None,
        method: Callable[..., Any] | str | None = None,  # Filter methods can return various types
        distinct: bool = False,
        exclude: bool = False,
        **kwargs: Any,  # Field kwargs stored as extra (required, help_text, etc.)
    ) -> None: ...

class ModelMultipleChoiceFilter(QuerySetRequestMixin, MultipleChoiceFilter):
    field_class: type[ModelMultipleChoiceField]  # More specific than parent MultipleChoiceField
    def __init__(
        self,
        field_name: str | None = None,
        lookup_expr: str | None = None,
        *,
        # Inherited from QuerySetRequestMixin
        queryset: QuerySet[Any] | None = None,
        # Inherited from MultipleChoiceFilter
        distinct: bool = True,  # Overrides distinct default
        conjoined: bool = False,
        null_value: Any = ...,  # Multiple choice null values vary by implementation
        # Inherited from Filter
        label: StrOrPromise | None = None,
        method: Callable[..., Any] | str | None = None,  # Filter methods can return various types
        exclude: bool = False,
        **kwargs: Any,  # Field kwargs stored as extra (required, help_text, etc.)
    ) -> None: ...

class NumberFilter(Filter):
    field_class: type[forms.DecimalField]
    def get_max_validator(self) -> Any: ...  # Validator can be various Django validator types
    @property
    def field(self) -> Field: ...

class NumericRangeFilter(Filter):
    field_class: type[RangeField]
    lookup_expr: str
    def filter(self, qs: QuerySet[Any], value: Any) -> QuerySet[Any]: ...

class RangeFilter(Filter):
    field_class: type[RangeField]
    lookup_expr: str
    def filter(self, qs: QuerySet[Any], value: Any) -> QuerySet[Any]: ...

class DateRangeFilter(ChoiceFilter):
    choices: list[tuple[str, str]] | None
    filters: dict[str, Filter] | None
    def __init__(
        self,
        choices: list[tuple[str, str]] | None = None,
        filters: dict[str, Filter] | None = None,
        field_name: str | None = None,
        lookup_expr: str | None = None,
        *,
        # Inherited from ChoiceFilter
        null_value: Any = ...,  # Null value can be any type (None, empty string, etc.)
        # Inherited from Filter
        label: StrOrPromise | None = None,
        method: Callable[..., Any] | str | None = None,  # Filter methods can return various types
        distinct: bool = False,
        exclude: bool = False,
        **kwargs: Any,  # Field kwargs stored as extra (required, help_text, etc.)
    ) -> None: ...  # Uses args/kwargs for choice and filter configuration
    def filter(self, qs: QuerySet[Any], value: Any) -> QuerySet[Any]: ...

class DateFromToRangeFilter(RangeFilter):
    field_class: type[DateRangeField]

class DateTimeFromToRangeFilter(RangeFilter):
    field_class: type[DateTimeRangeField]

class IsoDateTimeFromToRangeFilter(RangeFilter):
    field_class: type[IsoDateTimeRangeField]

class TimeRangeFilter(RangeFilter):
    field_class: type[TimeRangeField]

class AllValuesFilter(ChoiceFilter):
    @property
    def field(self) -> Field: ...

class AllValuesMultipleFilter(MultipleChoiceFilter):
    @property
    def field(self) -> Field: ...

class BaseCSVFilter(Filter):
    base_field_class: type[BaseCSVField] = ...
    field_class: type[Any]  # Base class for CSV-based filters

class BaseInFilter(BaseCSVFilter): ...

class BaseRangeFilter(BaseCSVFilter):
    base_field_class: type[BaseRangeField] = ...

class LookupChoiceFilter(Filter):
    field_class: type[forms.CharField]
    outer_class: type[LookupChoiceField] = ...
    empty_label: StrOrPromise | None
    lookup_choices: list[tuple[str, StrOrPromise]] | None
    def __init__(
        self,
        field_name: str | None = None,
        lookup_choices: list[tuple[str, StrOrPromise]] | None = None,
        field_class: type[Field] | None = None,
        *,
        empty_label: StrOrPromise = ...,
        # Inherited from Filter
        label: StrOrPromise | None = None,
        method: Callable[..., Any] | str | None = None,  # Filter methods can return various types
        distinct: bool = False,
        exclude: bool = False,
        **kwargs: Any,  # Field kwargs stored as extra (required, help_text, etc.)
    ) -> None: ...
    @classmethod
    def normalize_lookup(cls, lookup: str | tuple[str, StrOrPromise]) -> tuple[str, StrOrPromise]: ...
    def get_lookup_choices(self) -> list[tuple[str, StrOrPromise]]: ...
    @property
    def field(self) -> Field: ...
    lookup_expr: str
    def filter(self, qs: QuerySet[Any], lookup: Lookup) -> QuerySet[Any]: ...

class OrderingFilter(BaseCSVFilter, ChoiceFilter):
    # Inherits CSV field behavior for comma-separated ordering.
    # BaseCSVFilter constructs a custom ConcreteCSVField class that derives
    # from BaseCSVField.
    field_class: type[BaseCSVField]
    descending_fmt: str
    param_map: dict[str, str] | None
    def __init__(
        self,
        field_name: str | None = None,
        lookup_expr: str | None = None,
        *,
        fields: dict[str, str] | Iterable[str] | Iterable[tuple[str, str]] = ...,
        field_labels: dict[str, StrOrPromise] = ...,
        # Inherited from ChoiceFilter
        null_value: Any = ...,  # Null value can be any type (None, empty string, etc.)
        # Inherited from Filter
        label: StrOrPromise | None = None,
        method: Callable[..., Any] | str | None = None,  # Filter methods can return various types
        distinct: bool = False,
        exclude: bool = False,
        **kwargs: Any,  # Field kwargs stored as extra (required, help_text, etc.)
    ) -> None: ...
    def get_ordering_value(self, param: str) -> str: ...
    def filter(self, qs: QuerySet[Any], value: Any) -> QuerySet[Any]: ...
    @classmethod
    def normalize_fields(cls, fields: Any) -> list[str]: ...
    def build_choices(self, fields: Any, labels: dict[str, StrOrPromise] | None) -> list[tuple[str, str]]: ...

class FilterMethod:
    f: Filter
    def __init__(self, filter_instance: Filter) -> None: ...
    def __call__(self, qs: QuerySet[Any], value: Any) -> QuerySet[Any]: ...
    @property
    def method(self) -> Callable[..., Any]: ...



from _typeshed import SupportsKeysAndGetItem
from _weakref import getweakrefcount as getweakrefcount, getweakrefs as getweakrefs, proxy as proxy
from _weakrefset import WeakSet as WeakSet
from collections.abc import Callable, Iterable, Iterator, Mapping, MutableMapping
from types import GenericAlias
from typing import Any, ClassVar, Generic, TypeVar, final, overload
from typing_extensions import ParamSpec, Self, disjoint_base

__all__ = [
    "ref",
    "proxy",
    "getweakrefcount",
    "getweakrefs",
    "WeakKeyDictionary",
    "ReferenceType",
    "ProxyType",
    "CallableProxyType",
    "ProxyTypes",
    "WeakValueDictionary",
    "WeakSet",
    "WeakMethod",
    "finalize",
]

_T = TypeVar("_T")
_T1 = TypeVar("_T1")
_T2 = TypeVar("_T2")
_KT = TypeVar("_KT")
_VT = TypeVar("_VT")
_CallableT = TypeVar("_CallableT", bound=Callable[..., Any])
_P = ParamSpec("_P")

ProxyTypes: tuple[type[Any], ...]

# These classes are implemented in C and imported from _weakref at runtime. However,
# they consider themselves to live in the weakref module for sys.version_info >= (3, 11),
# so defining their stubs here means we match their __module__ value.
# Prior to 3.11 they did not declare a module for themselves and ended up looking like they
# came from the builtin module at runtime, which was just wrong, and we won't attempt to
# duplicate that.

@final
class CallableProxyType(Generic[_CallableT]):  # "weakcallableproxy"
    def __eq__(self, value: object, /) -> bool: ...
    def __getattr__(self, attr: str) -> Any: ...
    __call__: _CallableT
    __hash__: ClassVar[None]  # type: ignore[assignment]

@final
class ProxyType(Generic[_T]):  # "weakproxy"
    def __eq__(self, value: object, /) -> bool: ...
    def __getattr__(self, attr: str) -> Any: ...
    __hash__: ClassVar[None]  # type: ignore[assignment]

@disjoint_base
class ReferenceType(Generic[_T]):  # "weakref"
    __callback__: Callable[[Self], Any]
    def __new__(cls, o: _T, callback: Callable[[Self], Any] | None = ..., /) -> Self: ...
    def __call__(self) -> _T | None: ...
    def __eq__(self, value: object, /) -> bool: ...
    def __hash__(self) -> int: ...
    def __class_getitem__(cls, item: Any, /) -> GenericAlias: ...

ref = ReferenceType

# everything below here is implemented in weakref.py

class WeakMethod(ref[_CallableT]):
    __slots__ = ("_func_ref", "_meth_type", "_alive", "__weakref__")
    def __new__(cls, meth: _CallableT, callback: Callable[[Self], Any] | None = None) -> Self: ...
    def __call__(self) -> _CallableT | None: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __hash__(self) -> int: ...

class WeakValueDictionary(MutableMapping[_KT, _VT]):
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(
        self: WeakValueDictionary[_KT, _VT],  # pyright: ignore[reportInvalidTypeVarUse]  #11780
        other: Mapping[_KT, _VT] | Iterable[tuple[_KT, _VT]],
        /,
    ) -> None: ...
    @overload
    def __init__(
        self: WeakValueDictionary[str, _VT],  # pyright: ignore[reportInvalidTypeVarUse]  #11780
        other: Mapping[str, _VT] | Iterable[tuple[str, _VT]] = (),
        /,
        **kwargs: _VT,
    ) -> None: ...
    def __len__(self) -> int: ...
    def __getitem__(self, key: _KT) -> _VT: ...
    def __setitem__(self, key: _KT, value: _VT) -> None: ...
    def __delitem__(self, key: _KT) -> None: ...
    def __contains__(self, key: object) -> bool: ...
    def __iter__(self) -> Iterator[_KT]: ...
    def copy(self) -> WeakValueDictionary[_KT, _VT]: ...
    __copy__ = copy
    def __deepcopy__(self, memo: Any) -> Self: ...
    @overload
    def get(self, key: _KT, default: None = None) -> _VT | None: ...
    @overload
    def get(self, key: _KT, default: _VT) -> _VT: ...
    @overload
    def get(self, key: _KT, default: _T) -> _VT | _T: ...
    # These are incompatible with Mapping
    def keys(self) -> Iterator[_KT]: ...  # type: ignore[override]
    def values(self) -> Iterator[_VT]: ...  # type: ignore[override]
    def items(self) -> Iterator[tuple[_KT, _VT]]: ...  # type: ignore[override]
    def itervaluerefs(self) -> Iterator[KeyedRef[_KT, _VT]]: ...
    def valuerefs(self) -> list[KeyedRef[_KT, _VT]]: ...
    def setdefault(self, key: _KT, default: _VT) -> _VT: ...
    @overload
    def pop(self, key: _KT) -> _VT: ...
    @overload
    def pop(self, key: _KT, default: _VT) -> _VT: ...
    @overload
    def pop(self, key: _KT, default: _T) -> _VT | _T: ...
    @overload
    def update(self, other: SupportsKeysAndGetItem[_KT, _VT], /, **kwargs: _VT) -> None: ...
    @overload
    def update(self, other: Iterable[tuple[_KT, _VT]], /, **kwargs: _VT) -> None: ...
    @overload
    def update(self, other: None = None, /, **kwargs: _VT) -> None: ...
    def __or__(self, other: Mapping[_T1, _T2]) -> WeakValueDictionary[_KT | _T1, _VT | _T2]: ...
    def __ror__(self, other: Mapping[_T1, _T2]) -> WeakValueDictionary[_KT | _T1, _VT | _T2]: ...
    # WeakValueDictionary.__ior__ should be kept roughly in line with MutableMapping.update()
    @overload  # type: ignore[misc]
    def __ior__(self, other: SupportsKeysAndGetItem[_KT, _VT]) -> Self: ...
    @overload
    def __ior__(self, other: Iterable[tuple[_KT, _VT]]) -> Self: ...

class KeyedRef(ref[_T], Generic[_KT, _T]):
    __slots__ = ("key",)
    key: _KT
    def __new__(type, ob: _T, callback: Callable[[Self], Any], key: _KT) -> Self: ...
    def __init__(self, ob: _T, callback: Callable[[Self], Any], key: _KT) -> None: ...

class WeakKeyDictionary(MutableMapping[_KT, _VT]):
    @overload
    def __init__(self, dict: None = None) -> None: ...
    @overload
    def __init__(self, dict: Mapping[_KT, _VT] | Iterable[tuple[_KT, _VT]]) -> None: ...
    def __len__(self) -> int: ...
    def __getitem__(self, key: _KT) -> _VT: ...
    def __setitem__(self, key: _KT, value: _VT) -> None: ...
    def __delitem__(self, key: _KT) -> None: ...
    def __contains__(self, key: object) -> bool: ...
    def __iter__(self) -> Iterator[_KT]: ...
    def copy(self) -> WeakKeyDictionary[_KT, _VT]: ...
    __copy__ = copy
    def __deepcopy__(self, memo: Any) -> Self: ...
    @overload
    def get(self, key: _KT, default: None = None) -> _VT | None: ...
    @overload
    def get(self, key: _KT, default: _VT) -> _VT: ...
    @overload
    def get(self, key: _KT, default: _T) -> _VT | _T: ...
    # These are incompatible with Mapping
    def keys(self) -> Iterator[_KT]: ...  # type: ignore[override]
    def values(self) -> Iterator[_VT]: ...  # type: ignore[override]
    def items(self) -> Iterator[tuple[_KT, _VT]]: ...  # type: ignore[override]
    def keyrefs(self) -> list[ref[_KT]]: ...
    # Keep WeakKeyDictionary.setdefault in line with MutableMapping.setdefault, modulo positional-only differences
    @overload
    def setdefault(self: WeakKeyDictionary[_KT, _VT | None], key: _KT, default: None = None) -> _VT: ...
    @overload
    def setdefault(self, key: _KT, default: _VT) -> _VT: ...
    @overload
    def pop(self, key: _KT) -> _VT: ...
    @overload
    def pop(self, key: _KT, default: _VT) -> _VT: ...
    @overload
    def pop(self, key: _KT, default: _T) -> _VT | _T: ...
    @overload
    def update(self, dict: SupportsKeysAndGetItem[_KT, _VT], /, **kwargs: _VT) -> None: ...
    @overload
    def update(self, dict: Iterable[tuple[_KT, _VT]], /, **kwargs: _VT) -> None: ...
    @overload
    def update(self, dict: None = None, /, **kwargs: _VT) -> None: ...
    def __or__(self, other: Mapping[_T1, _T2]) -> WeakKeyDictionary[_KT | _T1, _VT | _T2]: ...
    def __ror__(self, other: Mapping[_T1, _T2]) -> WeakKeyDictionary[_KT | _T1, _VT | _T2]: ...
    # WeakKeyDictionary.__ior__ should be kept roughly in line with MutableMapping.update()
    @overload  # type: ignore[misc]
    def __ior__(self, other: SupportsKeysAndGetItem[_KT, _VT]) -> Self: ...
    @overload
    def __ior__(self, other: Iterable[tuple[_KT, _VT]]) -> Self: ...

class finalize(Generic[_P, _T]):
    __slots__ = ()
    def __init__(self, obj: _T, func: Callable[_P, Any], /, *args: _P.args, **kwargs: _P.kwargs) -> None: ...
    def __call__(self, _: Any = None) -> Any | None: ...
    def detach(self) -> tuple[_T, Callable[_P, Any], tuple[Any, ...], dict[str, Any]] | None: ...
    def peek(self) -> tuple[_T, Callable[_P, Any], tuple[Any, ...], dict[str, Any]] | None: ...
    @property
    def alive(self) -> bool: ...
    atexit: bool



from typing import Final

XK_hebrew_doublelowline: Final = 0xCDF
XK_hebrew_aleph: Final = 0xCE0
XK_hebrew_bet: Final = 0xCE1
XK_hebrew_beth: Final = 0xCE1
XK_hebrew_gimel: Final = 0xCE2
XK_hebrew_gimmel: Final = 0xCE2
XK_hebrew_dalet: Final = 0xCE3
XK_hebrew_daleth: Final = 0xCE3
XK_hebrew_he: Final = 0xCE4
XK_hebrew_waw: Final = 0xCE5
XK_hebrew_zain: Final = 0xCE6
XK_hebrew_zayin: Final = 0xCE6
XK_hebrew_chet: Final = 0xCE7
XK_hebrew_het: Final = 0xCE7
XK_hebrew_tet: Final = 0xCE8
XK_hebrew_teth: Final = 0xCE8
XK_hebrew_yod: Final = 0xCE9
XK_hebrew_finalkaph: Final = 0xCEA
XK_hebrew_kaph: Final = 0xCEB
XK_hebrew_lamed: Final = 0xCEC
XK_hebrew_finalmem: Final = 0xCED
XK_hebrew_mem: Final = 0xCEE
XK_hebrew_finalnun: Final = 0xCEF
XK_hebrew_nun: Final = 0xCF0
XK_hebrew_samech: Final = 0xCF1
XK_hebrew_samekh: Final = 0xCF1
XK_hebrew_ayin: Final = 0xCF2
XK_hebrew_finalpe: Final = 0xCF3
XK_hebrew_pe: Final = 0xCF4
XK_hebrew_finalzade: Final = 0xCF5
XK_hebrew_finalzadi: Final = 0xCF5
XK_hebrew_zade: Final = 0xCF6
XK_hebrew_zadi: Final = 0xCF6
XK_hebrew_qoph: Final = 0xCF7
XK_hebrew_kuf: Final = 0xCF7
XK_hebrew_resh: Final = 0xCF8
XK_hebrew_shin: Final = 0xCF9
XK_hebrew_taw: Final = 0xCFA
XK_hebrew_taf: Final = 0xCFA
XK_Hebrew_switch: Final = 0xFF7E



from _typeshed import Incomplete

from networkx.classes.graph import Graph, _Node
from networkx.utils.backends import _dispatchable

__all__ = ["current_flow_closeness_centrality", "information_centrality"]

@_dispatchable
def current_flow_closeness_centrality(
    G: Graph[_Node], weight: str | None = None, dtype: type = ..., solver: str = "lu"
) -> dict[Incomplete, float]: ...

information_centrality = current_flow_closeness_centrality



from _typeshed import Incomplete, ReadableBuffer, StrOrBytesPath
from typing import Final, NamedTuple
from typing_extensions import Self
from weakref import WeakKeyDictionary

from reportlab.pdfbase import pdfdoc, pdfmetrics

__version__: Final[str]

class TTFError(pdfdoc.PDFError): ...

def SUBSETN(n, table: ReadableBuffer | None = ...) -> bytes: ...
def makeToUnicodeCMap(fontname: str, subset) -> str: ...
def splice(stream, offset, value): ...

GF_ARG_1_AND_2_ARE_WORDS: Final = 1
GF_ARGS_ARE_XY_VALUES: Final = 2
GF_ROUND_XY_TO_GRID: Final = 4
GF_WE_HAVE_A_SCALE: Final = 8
GF_RESERVED: Final = 16
GF_MORE_COMPONENTS: Final = 32
GF_WE_HAVE_AN_X_AND_Y_SCALE: Final = 64
GF_WE_HAVE_A_TWO_BY_TWO: Final = 128
GF_WE_HAVE_INSTRUCTIONS: Final = 256
GF_USE_MY_METRICS: Final = 512
GF_OVERLAP_COMPOUND: Final = 1024
GF_SCALED_COMPONENT_OFFSET: Final = 2048
GF_UNSCALED_COMPONENT_OFFSET: Final = 4096

def TTFOpenFile(fn: StrOrBytesPath) -> tuple[StrOrBytesPath,]: ...

class TTFontParser:
    ttfVersions: Incomplete
    ttcVersions: Incomplete
    fileKind: str
    validate: Incomplete
    subfontNameX: bytes
    def __init__(self, file, validate: int = 0, subfontIndex: int = 0) -> None: ...
    ttcVersion: Incomplete
    numSubfonts: Incomplete
    subfontOffsets: Incomplete
    def readTTCHeader(self) -> None: ...
    def getSubfont(self, subfontIndex) -> None: ...
    numTables: Incomplete
    searchRange: Incomplete
    entrySelector: Incomplete
    rangeShift: Incomplete
    table: Incomplete
    tables: Incomplete
    def readTableDirectory(self) -> None: ...
    version: Incomplete
    def readHeader(self): ...
    filename: Incomplete
    def readFile(self, f) -> None: ...
    def checksumTables(self) -> None: ...
    def checksumFile(self) -> None: ...
    def get_table_pos(self, tag): ...
    def seek(self, pos) -> None: ...
    def skip(self, delta) -> None: ...
    def seek_table(self, tag, offset_in_table: int = 0): ...
    def read_tag(self): ...
    def get_chunk(self, pos, length): ...
    def read_uint8(self): ...
    def read_ushort(self): ...
    def read_ulong(self): ...
    def read_short(self): ...
    def get_ushort(self, pos): ...
    def get_ulong(self, pos): ...
    def get_table(self, tag): ...

class TTFontMaker:
    tables: Incomplete
    def __init__(self) -> None: ...
    def add(self, tag, data) -> None: ...
    def makeStream(self): ...

class CMapFmt2SubHeader(NamedTuple):
    firstCode: Incomplete
    entryCount: Incomplete
    idDelta: Incomplete
    idRangeOffset: Incomplete

class TTFNameBytes(bytes):
    ustr: Incomplete
    def __new__(cls, b, enc: str = "utf8"): ...

class TTFontFile(TTFontParser):
    def __init__(self, file, charInfo: int = 1, validate: int = 0, subfontIndex: int = 0) -> None: ...
    name: Incomplete
    familyName: Incomplete
    styleName: Incomplete
    fullName: Incomplete
    uniqueFontID: Incomplete
    fontRevision: Incomplete
    unitsPerEm: Incomplete
    bbox: Incomplete
    ascent: Incomplete
    descent: Incomplete
    capHeight: Incomplete
    stemV: Incomplete
    italicAngle: Incomplete
    underlinePosition: Incomplete
    underlineThickness: Incomplete
    flags: Incomplete
    numGlyphs: Incomplete
    charToGlyph: Incomplete
    defaultWidth: Incomplete
    charWidths: Incomplete
    hmetrics: Incomplete
    glyphPos: Incomplete
    def extractInfo(self, charInfo: int = 1): ...
    def makeSubset(self, subset): ...

FF_FIXED: Incomplete
FF_SERIF: Incomplete
FF_SYMBOLIC: Incomplete
FF_SCRIPT: Incomplete
FF_NONSYMBOLIC: Incomplete
FF_ITALIC: Incomplete
FF_ALLCAP: Incomplete
FF_SMALLCAP: Incomplete
FF_FORCEBOLD: Incomplete

class TTFontFace(TTFontFile, pdfmetrics.TypeFace):
    def __init__(self, filename, validate: int = 0, subfontIndex: int = 0) -> None: ...
    def getCharWidth(self, code): ...
    def addSubsetObjects(self, doc, fontname, subset): ...

class TTEncoding:
    name: str
    def __init__(self) -> None: ...

class TTFont:
    class State:
        namePrefix: str
        nextCode: int
        internalName: Incomplete
        frozen: int
        subsets: Incomplete
        def __init__(self, asciiReadable=None, ttf=None) -> None: ...

    fontName: str
    face: TTFontFace
    encoding: TTEncoding
    state: WeakKeyDictionary[Incomplete, State]
    def __init__(
        self,
        name: str,
        filename,
        validate: int = 0,
        subfontIndex: int = 0,
        asciiReadable: int | None = None,
        shapable: bool = True,
    ) -> None: ...
    def stringWidth(self, text, size, encoding: str = "utf8"): ...
    def splitString(self, text, doc, encoding: str = "utf-8"): ...
    def getSubsetInternalName(self, subset, doc): ...
    def addObjects(self, doc) -> None: ...
    @property
    def hbFace(self) -> Incomplete | None: ...
    def hbFont(self, fontSize: float = 10): ...
    @property
    def shapable(self) -> bool: ...
    @shapable.setter
    def shapable(self, v) -> None: ...
    def pdfScale(self, v): ...
    def unregister(self) -> None: ...

class ShapedFragWord(list[Incomplete]): ...

class ShapeData(NamedTuple):
    cluster: int
    x_advance: float
    y_advance: float
    x_offset: float
    y_offset: float
    width: float

class ShapedStr(str):
    def __new__(cls, s, shapeData: ShapeData | None = None) -> Self: ...
    def __radd__(self, other) -> Self: ...

def shapeStr(s: str, fontName: str, fontSize: float, force: bool = False): ...
def freshTTFont(ttfn, ttfpath, **kwds) -> TTFont: ...
def makeShapedFragWord(w, K: list[Incomplete] = [], V: list[Incomplete] = []) -> type[ShapedFragWord]: ...
def shapeFragWord(w, features=None, force: bool = False): ...



from __future__ import annotations

import unittest
from collections.abc import Iterator, Mapping
from datetime import datetime, timedelta
from decimal import Decimal
from fractions import Fraction
from typing import TypedDict, Union
from typing_extensions import assert_type
from unittest.mock import _ANY, ANY, AsyncMock, MagicMock, Mock, patch

case = unittest.TestCase()

###
# Tests for assertAlmostEqual
###

case.assertAlmostEqual(1, 2.4)
case.assertAlmostEqual(2.4, 2.41)
case.assertAlmostEqual(Fraction(49, 50), Fraction(48, 50))
case.assertAlmostEqual(3.14, complex(5, 6))
case.assertAlmostEqual(datetime(1999, 1, 2), datetime(1999, 1, 2, microsecond=1), delta=timedelta(hours=1))
case.assertAlmostEqual(datetime(1999, 1, 2), datetime(1999, 1, 2, microsecond=1), None, "foo", timedelta(hours=1))
case.assertAlmostEqual(Decimal("1.1"), Decimal("1.11"))
case.assertAlmostEqual(2.4, 2.41, places=8)
case.assertAlmostEqual(2.4, 2.41, delta=0.02)
case.assertAlmostEqual(2.4, 2.41, None, "foo", 0.02)

case.assertAlmostEqual(2.4, 2.41, places=9, delta=0.02)  # type: ignore
case.assertAlmostEqual("foo", "bar")  # type: ignore
case.assertAlmostEqual(datetime(1999, 1, 2), datetime(1999, 1, 2, microsecond=1))  # type: ignore
case.assertAlmostEqual(Decimal("0.4"), Fraction(1, 2))  # type: ignore
case.assertAlmostEqual(complex(2, 3), Decimal("0.9"))  # type: ignore

###
# Tests for assertNotAlmostEqual
###

case.assertAlmostEqual(1, 2.4)
case.assertNotAlmostEqual(Fraction(49, 50), Fraction(48, 50))
case.assertAlmostEqual(3.14, complex(5, 6))
case.assertNotAlmostEqual(datetime(1999, 1, 2), datetime(1999, 1, 2, microsecond=1), delta=timedelta(hours=1))
case.assertNotAlmostEqual(datetime(1999, 1, 2), datetime(1999, 1, 2, microsecond=1), None, "foo", timedelta(hours=1))

case.assertNotAlmostEqual(2.4, 2.41, places=9, delta=0.02)  # type: ignore
case.assertNotAlmostEqual("foo", "bar")  # type: ignore
case.assertNotAlmostEqual(datetime(1999, 1, 2), datetime(1999, 1, 2, microsecond=1))  # type: ignore
case.assertNotAlmostEqual(Decimal("0.4"), Fraction(1, 2))  # type: ignore
case.assertNotAlmostEqual(complex(2, 3), Decimal("0.9"))  # type: ignore

###
# Tests for assertGreater
###


class Spam:
    def __lt__(self, other: object) -> bool:
        return True


class Eggs:
    def __gt__(self, other: object) -> bool:
        return True


class Ham:
    def __lt__(self, other: Ham) -> bool:
        if not isinstance(other, Ham):
            return NotImplemented
        return True


class Bacon:
    def __gt__(self, other: Bacon) -> bool:
        if not isinstance(other, Bacon):
            return NotImplemented
        return True


case.assertGreater(5.8, 3)
case.assertGreater(Decimal("4.5"), Fraction(3, 2))
case.assertGreater(Fraction(3, 2), 0.9)
case.assertGreater(Eggs(), object())
case.assertGreater(object(), Spam())
case.assertGreater(Ham(), Ham())
case.assertGreater(Bacon(), Bacon())

case.assertGreater(object(), object())  # type: ignore
case.assertGreater(datetime(1999, 1, 2), 1)  # type: ignore
case.assertGreater(Spam(), Eggs())  # type: ignore
case.assertGreater(Ham(), Bacon())  # type: ignore
case.assertGreater(Bacon(), Ham())  # type: ignore


###
# Tests for assertDictEqual
###


class TD1(TypedDict):
    x: int
    y: str


class TD2(TypedDict):
    a: bool
    b: bool


class MyMapping(Mapping[str, int]):
    def __getitem__(self, __key: str) -> int:
        return 42

    def __iter__(self) -> Iterator[str]:
        return iter([])

    def __len__(self) -> int:
        return 0


td1: TD1 = {"x": 1, "y": "foo"}
td2: TD2 = {"a": True, "b": False}
m = MyMapping()

case.assertDictEqual({}, {})
case.assertDictEqual({"x": 1, "y": 2}, {"x": 1, "y": 2})
case.assertDictEqual({"x": 1, "y": "foo"}, {"y": "foo", "x": 1})
case.assertDictEqual({"x": 1}, {})
case.assertDictEqual({}, {"x": 1})
case.assertDictEqual({1: "x"}, {"y": 222})
case.assertDictEqual({1: "x"}, td1)
case.assertDictEqual(td1, {1: "x"})
case.assertDictEqual(td1, td2)

case.assertDictEqual(1, {})  # type: ignore
case.assertDictEqual({}, 1)  # type: ignore

# These should fail, but don't due to TypedDict limitations:
# case.assertDictEqual(m, {"": 0})  # xtype: ignore
# case.assertDictEqual({"": 0}, m)  # xtype: ignore

###
# Tests for mock.patch
###


@patch("sys.exit")
def f_default_new(i: int, mock: MagicMock) -> str:
    return "asdf"


@patch("sys.exit", new=42)
def f_explicit_new(i: int) -> str:
    return "asdf"


@patch("sys.exit", new_callable=lambda: 42)
def f_explicit_new_callable(i: int, new_callable_ret: int) -> str:
    return "asdf"


assert_type(f_default_new(1), str)
f_default_new("a")  # Not an error due to ParamSpec limitations
assert_type(f_explicit_new(1), str)
f_explicit_new("a")  # type: ignore[arg-type]
assert_type(f_explicit_new_callable(1), str)
f_explicit_new_callable("a")  # Same as default new


@patch("sys.exit", new=Mock())
class TestXYZ(unittest.TestCase):
    attr: int = 5

    @staticmethod
    def method() -> int:
        return 123


assert_type(TestXYZ.attr, int)
assert_type(TestXYZ.method(), int)


with patch("sys.exit") as default_new_enter:
    assert_type(default_new_enter, Union[MagicMock, AsyncMock])

with patch("sys.exit", new=42) as explicit_new_enter:
    assert_type(explicit_new_enter, int)

with patch("sys.exit", new_callable=lambda: 42) as explicit_new_callable_enter:
    assert_type(explicit_new_callable_enter, int)


###
# Tests for mock.patch.object
###


@patch.object(Decimal, "exp")
def obj_f_default_new(i: int, mock: MagicMock) -> str:
    return "asdf"


@patch.object(Decimal, "exp", new=42)
def obj_f_explicit_new(i: int) -> str:
    return "asdf"


@patch.object(Decimal, "exp", new_callable=lambda: 42)
def obj_f_explicit_new_callable(i: int, new_callable_ret: int) -> str:
    return "asdf"


assert_type(obj_f_default_new(1), str)
obj_f_default_new("a")  # Not an error due to ParamSpec limitations
assert_type(obj_f_explicit_new(1), str)
obj_f_explicit_new("a")  # type: ignore[arg-type]
assert_type(obj_f_explicit_new_callable(1), str)
obj_f_explicit_new_callable("a")  # Same as default new


with patch.object(Decimal, "exp") as obj_default_new_enter:
    assert_type(obj_default_new_enter, Union[MagicMock, AsyncMock])

with patch.object(Decimal, "exp", new=42) as obj_explicit_new_enter:
    assert_type(obj_explicit_new_enter, int)

with patch.object(Decimal, "exp", new_callable=lambda: 42) as obj_explicit_new_callable_enter:
    assert_type(obj_explicit_new_callable_enter, int)


###
# Tests for mock.ANY
###


assert_type(ANY, _ANY)  # Make sure ANY has runtime type


# Regression tests. See https://github.com/python/typeshed/issues/14701
class TD(TypedDict):
    x: str
    y: float


td: TD = {"x": "1", "y": ANY}


def test_as_param(x: str) -> None: ...


test_as_param(ANY)


def test_as_return_value(x: int) -> TD:
    return {"x": str(x), "y": ANY}



from collections.abc import Mapping
from http.cookiejar import CookieJar
from typing_extensions import TypeAlias

from .models import _JSON, Response
from .sessions import _Auth, _Cert, _Data, _Files, _HooksInput, _Params, _TextMapping, _Timeout, _Verify

_HeadersMapping: TypeAlias = Mapping[str, str | bytes | None]

def request(
    method: str | bytes,
    url: str | bytes,
    *,
    params: _Params | None = ...,
    data: _Data | None = ...,
    headers: _HeadersMapping | None = ...,
    cookies: CookieJar | _TextMapping | None = ...,
    files: _Files | None = ...,
    auth: _Auth | None = ...,
    timeout: _Timeout | None = ...,
    allow_redirects: bool = ...,
    proxies: _TextMapping | None = ...,
    hooks: _HooksInput | None = ...,
    stream: bool | None = ...,
    verify: _Verify | None = ...,
    cert: _Cert | None = ...,
    json: _JSON | None = None,
) -> Response: ...
def get(
    url: str | bytes,
    params: _Params | None = None,
    *,
    data: _Data | None = ...,
    headers: _HeadersMapping | None = ...,
    cookies: CookieJar | _TextMapping | None = ...,
    files: _Files | None = ...,
    auth: _Auth | None = ...,
    timeout: _Timeout | None = ...,
    allow_redirects: bool = ...,
    proxies: _TextMapping | None = ...,
    hooks: _HooksInput | None = ...,
    stream: bool | None = ...,
    verify: _Verify | None = ...,
    cert: _Cert | None = ...,
    json: _JSON | None = None,
) -> Response: ...
def options(
    url: str | bytes,
    *,
    params: _Params | None = ...,
    data: _Data | None = ...,
    headers: _HeadersMapping | None = ...,
    cookies: CookieJar | _TextMapping | None = ...,
    files: _Files | None = ...,
    auth: _Auth | None = ...,
    timeout: _Timeout | None = ...,
    allow_redirects: bool = ...,
    proxies: _TextMapping | None = ...,
    hooks: _HooksInput | None = ...,
    stream: bool | None = ...,
    verify: _Verify | None = ...,
    cert: _Cert | None = ...,
    json: _JSON | None = None,
) -> Response: ...
def head(
    url: str | bytes,
    *,
    params: _Params | None = ...,
    data: _Data | None = ...,
    headers: _HeadersMapping | None = ...,
    cookies: CookieJar | _TextMapping | None = ...,
    files: _Files | None = ...,
    auth: _Auth | None = ...,
    timeout: _Timeout | None = ...,
    allow_redirects: bool = ...,
    proxies: _TextMapping | None = ...,
    hooks: _HooksInput | None = ...,
    stream: bool | None = ...,
    verify: _Verify | None = ...,
    cert: _Cert | None = ...,
    json: _JSON | None = None,
) -> Response: ...
def post(
    url: str | bytes,
    data: _Data | None = None,
    json: _JSON | None = None,
    *,
    params: _Params | None = ...,
    headers: _HeadersMapping | None = ...,
    cookies: CookieJar | _TextMapping | None = ...,
    files: _Files | None = ...,
    auth: _Auth | None = ...,
    timeout: _Timeout | None = ...,
    allow_redirects: bool = ...,
    proxies: _TextMapping | None = ...,
    hooks: _HooksInput | None = ...,
    stream: bool | None = ...,
    verify: _Verify | None = ...,
    cert: _Cert | None = ...,
) -> Response: ...
def put(
    url: str | bytes,
    data: _Data | None = None,
    *,
    params: _Params | None = ...,
    headers: _HeadersMapping | None = ...,
    cookies: CookieJar | _TextMapping | None = ...,
    files: _Files | None = ...,
    auth: _Auth | None = ...,
    timeout: _Timeout | None = ...,
    allow_redirects: bool = ...,
    proxies: _TextMapping | None = ...,
    hooks: _HooksInput | None = ...,
    stream: bool | None = ...,
    verify: _Verify | None = ...,
    cert: _Cert | None = ...,
    json: _JSON | None = None,
) -> Response: ...
def patch(
    url: str | bytes,
    data: _Data | None = None,
    *,
    params: _Params | None = ...,
    headers: _HeadersMapping | None = ...,
    cookies: CookieJar | _TextMapping | None = ...,
    files: _Files | None = ...,
    auth: _Auth | None = ...,
    timeout: _Timeout | None = ...,
    allow_redirects: bool = ...,
    proxies: _TextMapping | None = ...,
    hooks: _HooksInput | None = ...,
    stream: bool | None = ...,
    verify: _Verify | None = ...,
    cert: _Cert | None = ...,
    json: _JSON | None = None,
) -> Response: ...
def delete(
    url: str | bytes,
    *,
    params: _Params | None = ...,
    data: _Data | None = ...,
    headers: _HeadersMapping | None = ...,
    cookies: CookieJar | _TextMapping | None = ...,
    files: _Files | None = ...,
    auth: _Auth | None = ...,
    timeout: _Timeout | None = ...,
    allow_redirects: bool = ...,
    proxies: _TextMapping | None = ...,
    hooks: _HooksInput | None = ...,
    stream: bool | None = ...,
    verify: _Verify | None = ...,
    cert: _Cert | None = ...,
    json: _JSON | None = None,
) -> Response: ...



from _typeshed import Incomplete
from typing import Final

__version__: Final[str]

def replace(src, sep, rep): ...

keywordsList: Incomplete
commentPat: str
pat: str
quotePat: Incomplete
tripleQuotePat: Incomplete
nonKeyPat: str
keyPat: Incomplete
matchPat: Incomplete
matchRE: Incomplete
idKeyPat: str
idRE: Incomplete

def fontify(pytext, searchfrom: int = 0, searchto=None): ...
def test(path) -> None: ...



from collections.abc import Sequence
from typing import Any, Generic, TypeVar

from django.db.models import Model

from .instance_loaders import BaseInstanceLoader

_ModelT = TypeVar("_ModelT", bound=Model)

class ResourceOptions(Generic[_ModelT]):
    model: _ModelT | str
    fields: Sequence[str] | None
    exclude: Sequence[str] | None
    instance_loader_class: type[BaseInstanceLoader] | None
    import_id_fields: Sequence[str]
    export_order: Sequence[str] | None
    import_order: Sequence[str] | None
    widgets: dict[str, Any] | None
    use_transactions: bool | None
    skip_unchanged: bool
    report_skipped: bool
    clean_model_instances: bool
    chunk_size: int | None
    skip_diff: bool
    skip_html_diff: bool
    use_bulk: bool
    batch_size: int
    force_init_instance: bool
    using_db: str | None
    store_row_values: bool
    store_instance: bool
    use_natural_foreign_keys: bool



from _typeshed import Incomplete

lock: Incomplete
CDEF_SOURCE_STRING: str

class Parser:
    def __init__(self) -> None: ...
    def convert_pycparser_error(self, e, csource) -> None: ...
    def parse(self, csource, override: bool = False, packed: bool = False, pack=None, dllexport: bool = False) -> None: ...
    def parse_type(self, cdecl): ...
    def parse_type_and_quals(self, cdecl): ...
    def include(self, other) -> None: ...



from _typeshed import Incomplete
from typing import Final

import grpc

GRPC_GENERATED_VERSION: Final[str]
GRPC_VERSION: Final[str]

class ChannelzStub:
    GetTopChannels: Incomplete
    GetServers: Incomplete
    GetServer: Incomplete
    GetServerSockets: Incomplete
    GetChannel: Incomplete
    GetSubchannel: Incomplete
    GetSocket: Incomplete
    def __init__(self, channel: grpc.Channel): ...

class ChannelzServicer:
    def GetTopChannels(self, request, context): ...
    def GetServers(self, request, context): ...
    def GetServer(self, request, context): ...
    def GetServerSockets(self, request, context): ...
    def GetChannel(self, request, context): ...
    def GetSubchannel(self, request, context): ...
    def GetSocket(self, request, context): ...

def add_ChannelzServicer_to_server(servicer, server) -> None: ...

class Channelz:
    @staticmethod
    def GetTopChannels(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ): ...
    @staticmethod
    def GetServers(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ): ...
    @staticmethod
    def GetServer(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ): ...
    @staticmethod
    def GetServerSockets(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ): ...
    @staticmethod
    def GetChannel(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ): ...
    @staticmethod
    def GetSubchannel(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ): ...
    @staticmethod
    def GetSocket(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ): ...



"""Jac Program module.

This module provides the JacProgram class which holds the state of a compiled
Jac program. The actual compilation is performed by JacCompiler.
"""

from __future__ import annotations

import types
from threading import Event
from typing import TYPE_CHECKING

import jaclang.pycore.unitree as uni
from jaclang.pycore.bccache import BytecodeCache, get_bytecode_cache
from jaclang.pycore.passes import Alert

if TYPE_CHECKING:
    from jaclang.compiler.type_system.type_evaluator import TypeEvaluator
    from jaclang.pycore.compiler import JacCompiler
    from jaclang.pycore.passes import Transform


class JacProgram:
    """JacProgram holds the state of a compiled Jac program.

    This class contains:
    - mod: The ProgramModule containing all compiled modules in mod.hub
    - errors_had: List of compilation errors
    - warnings_had: List of compilation warnings
    - type_evaluator: Optional type evaluator for type checking
    - py_raise_map: Mapping for Python exception translation

    Compilation is performed by JacCompiler, which takes a JacProgram
    as a target to store compiled modules and track errors.
    """

    def __init__(
        self,
        main_mod: uni.ProgramModule | None = None,
        bytecode_cache: BytecodeCache | None = None,
    ) -> None:
        """Initialize the JacProgram object.

        Args:
            main_mod: Optional main module to initialize with.
            bytecode_cache: Optional custom bytecode cache. If None, uses default.
        """
        self.mod: uni.ProgramModule = main_mod if main_mod else uni.ProgramModule()
        self.py_raise_map: dict[str, str] = {}
        self.errors_had: list[Alert] = []
        self.warnings_had: list[Alert] = []
        self.type_evaluator: TypeEvaluator | None = None
        self._bytecode_cache: BytecodeCache = bytecode_cache or get_bytecode_cache()

    def _get_compiler(self) -> JacCompiler:
        """Get a JacCompiler instance configured with this program's cache."""
        from jaclang.pycore.compiler import JacCompiler

        return JacCompiler(bytecode_cache=self._bytecode_cache)

    def get_type_evaluator(self) -> TypeEvaluator:
        """Return the type evaluator, creating one if needed."""
        from jaclang.compiler.type_system.type_evaluator import TypeEvaluator

        if not self.type_evaluator:
            self.type_evaluator = TypeEvaluator(program=self)
        return self.type_evaluator

    def clear_type_system(self, clear_hub: bool = False) -> None:
        """Clear all type information from the program.

        This method resets the type evaluator and clears cached type information
        from all AST nodes. This is useful for test isolation when running multiple
        tests in the same process, as type information attached to AST nodes can
        persist in sys.modules and pollute subsequent tests.

        Args:
            clear_hub: If True, also clear all modules from mod.hub. Use with
                       caution as this removes all compiled modules.
        """
        # Clear the type evaluator (will be recreated lazily if needed)
        self.type_evaluator = None

        # Optionally clear the entire module hub (skip node traversal if clearing hub)
        if clear_hub:
            self.mod.hub.clear()
        else:
            # Clear .type attributes from all Expr nodes in all modules
            for mod in self.mod.hub.values():
                for node in mod.get_all_sub_nodes(uni.Expr, brute_force=True):
                    node.type = None

    def get_bytecode(
        self, full_target: str, minimal: bool = False
    ) -> types.CodeType | None:
        """Get the bytecode for a specific module.

        Args:
            full_target: The full path to the module file.
            minimal: If True, use minimal compilation (no JS/type analysis).
                     This avoids circular imports for bootstrap-critical modules.
        """
        return self._get_compiler().get_bytecode(full_target, self, minimal)

    def parse_str(
        self, source_str: str, file_path: str, cancel_token: Event | None = None
    ) -> uni.Module:
        """Parse source string into an AST module."""
        return self._get_compiler().parse_str(source_str, file_path, self, cancel_token)

    def compile(
        self,
        file_path: str,
        use_str: str | None = None,
        no_cgen: bool = False,
        type_check: bool = False,
        symtab_ir_only: bool = False,
        minimal: bool = False,
        cancel_token: Event | None = None,
    ) -> uni.Module:
        """Compile a Jac file into a module AST.

        Args:
            file_path: Path to the Jac file to compile.
            use_str: Optional source string to use instead of reading from file.
            no_cgen: If True, skip code generation entirely.
            type_check: If True, run type checking pass.
            symtab_ir_only: If True, only build symbol table (skip semantic analysis).
            minimal: If True, use minimal compilation mode (bytecode only, no JS).
                     This avoids circular imports for bootstrap-critical modules.
            cancel_token: Optional event to cancel compilation.
        """
        return self._get_compiler().compile(
            file_path=file_path,
            target_program=self,
            use_str=use_str,
            no_cgen=no_cgen,
            type_check=type_check,
            symtab_ir_only=symtab_ir_only,
            minimal=minimal,
            cancel_token=cancel_token,
        )

    def build(
        self, file_path: str, use_str: str | None = None, type_check: bool = False
    ) -> uni.Module:
        """Build a Jac file with import dependency resolution."""
        return self._get_compiler().build(
            file_path, self, use_str, type_check=type_check
        )

    def run_schedule(
        self,
        mod: uni.Module,
        passes: list[type[Transform[uni.Module, uni.Module]]],
        cancel_token: Event | None = None,
    ) -> None:
        """Run a schedule of passes on a module."""
        self._get_compiler().run_schedule(mod, self, passes, cancel_token)

    @staticmethod
    def jac_file_formatter(file_path: str, auto_lint: bool = False) -> JacProgram:
        """Format a Jac file and return the JacProgram.

        Args:
            file_path: Path to the Jac file to format.
            auto_lint: If True, apply auto-linting corrections before formatting.
        """
        from jaclang.pycore.compiler import JacCompiler

        return JacCompiler.jac_file_formatter(file_path, auto_lint)

    @staticmethod
    def jac_str_formatter(
        source_str: str, file_path: str, auto_lint: bool = False
    ) -> JacProgram:
        """Format a Jac string and return the JacProgram.

        Args:
            source_str: The Jac source code string to format.
            file_path: Path to use for error messages.
            auto_lint: If True, apply auto-linting corrections before formatting.
        """
        from jaclang.pycore.compiler import JacCompiler

        return JacCompiler.jac_str_formatter(source_str, file_path, auto_lint)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Defines the text format for including per-op API definition and
overrides for client language op code generators.
"""

import builtins
import collections.abc
import sys
import typing

import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import tensorflow.core.framework.attr_value_pb2

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class ApiDef(google.protobuf.message.Message):
    """Used to specify and override the default API & behavior in the
    generated code for client languages, from what you would get from
    the OpDef alone. There will be a set of ApiDefs that are common
    to all client languages, and another set per client language.
    The per-client-language ApiDefs will inherit values from the
    common ApiDefs which it can either replace or modify.

    We separate the API definition from the OpDef so we can evolve the
    API while remaining backwards compatible when interpreting old
    graphs.  Overrides go in an "api_def.pbtxt" file with a text-format
    ApiDefs message.

    WARNING: Be *very* careful changing the API for any existing op --
    you can change the semantics of existing code.  These changes may
    need to wait until a major release of TensorFlow to avoid breaking
    our compatibility promises.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _Visibility:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _VisibilityEnumTypeWrapper(
        google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[ApiDef._Visibility.ValueType], builtins.type
    ):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        DEFAULT_VISIBILITY: ApiDef._Visibility.ValueType  # 0
        """Normally this is "VISIBLE" unless you are inheriting a
        different value from another ApiDef.
        """
        VISIBLE: ApiDef._Visibility.ValueType  # 1
        """Publicly visible in the API."""
        SKIP: ApiDef._Visibility.ValueType  # 2
        """Do not include this op in the generated API. If visibility is
        set to 'SKIP', other fields are ignored for this op.
        """
        HIDDEN: ApiDef._Visibility.ValueType  # 3
        """Hide this op by putting it into an internal namespace (or whatever
        is appropriate in the target language).
        """

    class Visibility(_Visibility, metaclass=_VisibilityEnumTypeWrapper): ...
    DEFAULT_VISIBILITY: ApiDef.Visibility.ValueType  # 0
    """Normally this is "VISIBLE" unless you are inheriting a
    different value from another ApiDef.
    """
    VISIBLE: ApiDef.Visibility.ValueType  # 1
    """Publicly visible in the API."""
    SKIP: ApiDef.Visibility.ValueType  # 2
    """Do not include this op in the generated API. If visibility is
    set to 'SKIP', other fields are ignored for this op.
    """
    HIDDEN: ApiDef.Visibility.ValueType  # 3
    """Hide this op by putting it into an internal namespace (or whatever
    is appropriate in the target language).
    """

    @typing.final
    class Endpoint(google.protobuf.message.Message):
        """If you specify any endpoint, this will replace all of the
        inherited endpoints.  The first endpoint should be the
        "canonical" endpoint, and should not be deprecated (unless all
        endpoints are deprecated).
        """

        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        NAME_FIELD_NUMBER: builtins.int
        DEPRECATED_FIELD_NUMBER: builtins.int
        DEPRECATION_VERSION_FIELD_NUMBER: builtins.int
        name: builtins.str
        """Name should be either like "CamelCaseName" or
        "Package.CamelCaseName". Client-language-specific ApiDefs may
        use a snake_case convention instead of CamelCase.
        """
        deprecated: builtins.bool
        """Set if this endpoint is deprecated. If set to true, a message suggesting
        to use a non-deprecated endpoint instead will be printed. If all
        endpoints are deprecated, set deprecation_message in ApiDef instead.
        """
        deprecation_version: builtins.int
        """Major version when an endpoint will be deleted. For e.g. set this
        value to 2 if endpoint should be removed in TensorFlow 2.0 and
        deprecated in versions before that.
        """
        def __init__(
            self,
            *,
            name: builtins.str | None = ...,
            deprecated: builtins.bool | None = ...,
            deprecation_version: builtins.int | None = ...,
        ) -> None: ...
        def ClearField(
            self,
            field_name: typing.Literal[
                "deprecated", b"deprecated", "deprecation_version", b"deprecation_version", "name", b"name"
            ],
        ) -> None: ...

    @typing.final
    class Arg(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        NAME_FIELD_NUMBER: builtins.int
        RENAME_TO_FIELD_NUMBER: builtins.int
        DESCRIPTION_FIELD_NUMBER: builtins.int
        name: builtins.str
        rename_to: builtins.str
        """Change the name used to access this arg in the API from what
        is used in the GraphDef.  Note that these names in `backticks`
        will also be replaced in the summary & description fields.
        """
        description: builtins.str
        """Note: this will replace any inherited arg doc. There is no
        current way of modifying arg descriptions (other than replacing
        them entirely) as can be done with op descriptions.
        """
        def __init__(
            self, *, name: builtins.str | None = ..., rename_to: builtins.str | None = ..., description: builtins.str | None = ...
        ) -> None: ...
        def ClearField(
            self, field_name: typing.Literal["description", b"description", "name", b"name", "rename_to", b"rename_to"]
        ) -> None: ...

    @typing.final
    class Attr(google.protobuf.message.Message):
        """Description of the graph-construction-time configuration of this
        Op.  That is to say, this describes the attr fields that will
        be specified in the NodeDef.
        """

        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        NAME_FIELD_NUMBER: builtins.int
        RENAME_TO_FIELD_NUMBER: builtins.int
        DEFAULT_VALUE_FIELD_NUMBER: builtins.int
        DESCRIPTION_FIELD_NUMBER: builtins.int
        name: builtins.str
        rename_to: builtins.str
        """Change the name used to access this attr in the API from what
        is used in the GraphDef.  Note that these names in `backticks`
        will also be replaced in the summary & description fields.
        """
        description: builtins.str
        """Note: this will replace any inherited attr doc, there is no current
        way of modifying attr descriptions as can be done with op descriptions.
        """
        @property
        def default_value(self) -> tensorflow.core.framework.attr_value_pb2.AttrValue:
            """Specify a new default value to use for this attr.  This default
            will be used when creating new graphs, as opposed to the
            default in the OpDef, which will be used when interpreting old
            GraphDefs.
            """

        def __init__(
            self,
            *,
            name: builtins.str | None = ...,
            rename_to: builtins.str | None = ...,
            default_value: tensorflow.core.framework.attr_value_pb2.AttrValue | None = ...,
            description: builtins.str | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["default_value", b"default_value"]) -> builtins.bool: ...
        def ClearField(
            self,
            field_name: typing.Literal[
                "default_value", b"default_value", "description", b"description", "name", b"name", "rename_to", b"rename_to"
            ],
        ) -> None: ...

    GRAPH_OP_NAME_FIELD_NUMBER: builtins.int
    DEPRECATION_MESSAGE_FIELD_NUMBER: builtins.int
    DEPRECATION_VERSION_FIELD_NUMBER: builtins.int
    VISIBILITY_FIELD_NUMBER: builtins.int
    ENDPOINT_FIELD_NUMBER: builtins.int
    IN_ARG_FIELD_NUMBER: builtins.int
    OUT_ARG_FIELD_NUMBER: builtins.int
    ARG_ORDER_FIELD_NUMBER: builtins.int
    ATTR_FIELD_NUMBER: builtins.int
    SUMMARY_FIELD_NUMBER: builtins.int
    DESCRIPTION_FIELD_NUMBER: builtins.int
    DESCRIPTION_PREFIX_FIELD_NUMBER: builtins.int
    DESCRIPTION_SUFFIX_FIELD_NUMBER: builtins.int
    graph_op_name: builtins.str
    """Name of the op (in the OpDef) to specify the API for."""
    deprecation_message: builtins.str
    """If this op is deprecated, set deprecation message to the message
    that should be logged when this op is used.
    The message should indicate alternative op to use, if any.
    """
    deprecation_version: builtins.int
    """Major version when the op will be deleted. For e.g. set this
    value to 2 if op API should be removed in TensorFlow 2.0 and
    deprecated in versions before that.
    """
    visibility: global___ApiDef.Visibility.ValueType
    summary: builtins.str
    """One-line human-readable description of what the Op does."""
    description: builtins.str
    """Additional, longer human-readable description of what the Op does."""
    description_prefix: builtins.str
    """Modify an existing/inherited description by adding text to the beginning
    or end.
    """
    description_suffix: builtins.str
    @property
    def endpoint(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ApiDef.Endpoint]: ...
    @property
    def in_arg(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ApiDef.Arg]: ...
    @property
    def out_arg(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ApiDef.Arg]: ...
    @property
    def arg_order(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """List of original in_arg names to specify new argument order.
        Length of arg_order should be either empty to keep current order
        or match size of in_arg.
        """

    @property
    def attr(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ApiDef.Attr]: ...
    def __init__(
        self,
        *,
        graph_op_name: builtins.str | None = ...,
        deprecation_message: builtins.str | None = ...,
        deprecation_version: builtins.int | None = ...,
        visibility: global___ApiDef.Visibility.ValueType | None = ...,
        endpoint: collections.abc.Iterable[global___ApiDef.Endpoint] | None = ...,
        in_arg: collections.abc.Iterable[global___ApiDef.Arg] | None = ...,
        out_arg: collections.abc.Iterable[global___ApiDef.Arg] | None = ...,
        arg_order: collections.abc.Iterable[builtins.str] | None = ...,
        attr: collections.abc.Iterable[global___ApiDef.Attr] | None = ...,
        summary: builtins.str | None = ...,
        description: builtins.str | None = ...,
        description_prefix: builtins.str | None = ...,
        description_suffix: builtins.str | None = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing.Literal[
            "arg_order",
            b"arg_order",
            "attr",
            b"attr",
            "deprecation_message",
            b"deprecation_message",
            "deprecation_version",
            b"deprecation_version",
            "description",
            b"description",
            "description_prefix",
            b"description_prefix",
            "description_suffix",
            b"description_suffix",
            "endpoint",
            b"endpoint",
            "graph_op_name",
            b"graph_op_name",
            "in_arg",
            b"in_arg",
            "out_arg",
            b"out_arg",
            "summary",
            b"summary",
            "visibility",
            b"visibility",
        ],
    ) -> None: ...

global___ApiDef = ApiDef

@typing.final
class ApiDefs(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    OP_FIELD_NUMBER: builtins.int
    @property
    def op(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ApiDef]: ...
    def __init__(self, *, op: collections.abc.Iterable[global___ApiDef] | None = ...) -> None: ...
    def ClearField(self, field_name: typing.Literal["op", b"op"]) -> None: ...

global___ApiDefs = ApiDefs



from collections.abc import Hashable, Iterable

def eval(a: Iterable[Hashable], b: Iterable[Hashable]) -> int: ...
def distance(a: Iterable[Hashable], b: Iterable[Hashable]) -> int: ...
def eval_criterion(a: Iterable[Hashable], b: Iterable[Hashable], thr: int) -> bool: ...
def distance_le_than(a: Iterable[Hashable], b: Iterable[Hashable], thr: int) -> bool: ...

__all__ = ("eval", "distance", "eval_criterion", "distance_le_than")



from _typeshed.wsgi import StartResponse, WSGIApplication, WSGIEnvironment
from collections.abc import Callable, Iterable, Mapping
from typing import Any, Generic, overload, type_check_only
from typing_extensions import Concatenate, Never, ParamSpec, Self, TypeAlias, TypeVar

from webob.request import BaseRequest, Request
from webob.response import Response

__all__ = ["wsgify"]

_AnyResponse: TypeAlias = Response | WSGIApplication | str | None
_S = TypeVar("_S")
_AppT = TypeVar("_AppT", bound=WSGIApplication)
_AppT_contra = TypeVar("_AppT_contra", bound=WSGIApplication, contravariant=True)
_RequestT = TypeVar("_RequestT", bound=BaseRequest)
_RequestT_contra = TypeVar("_RequestT_contra", bound=BaseRequest, default=Request, contravariant=True)
_P = ParamSpec("_P")
_P2 = ParamSpec("_P2")

_RequestHandlerCallable: TypeAlias = Callable[Concatenate[_RequestT, _P], _AnyResponse]
_RequestHandlerMethod: TypeAlias = Callable[Concatenate[Any, _RequestT, _P], _AnyResponse]
_MiddlewareCallable: TypeAlias = Callable[Concatenate[_RequestT, _AppT, _P], _AnyResponse]
_MiddlewareMethod: TypeAlias = Callable[Concatenate[Any, _RequestT, _AppT, _P], _AnyResponse]
_RequestHandler: TypeAlias = _RequestHandlerCallable[_RequestT, _P] | _RequestHandlerMethod[_RequestT, _P]
_Middleware: TypeAlias = _MiddlewareCallable[_RequestT, _AppT, _P] | _MiddlewareMethod[_RequestT, _AppT, _P]

class wsgify(Generic[_P, _RequestT_contra]):
    RequestClass: type[_RequestT_contra]
    func: _RequestHandler[_RequestT_contra, _P] | None
    args: tuple[Any, ...]
    kwargs: dict[str, Any]
    middleware_wraps: WSGIApplication | None
    # NOTE: We disallow passing args/kwargs using this direct API, because
    #       we can't really make it work as a decorator this way, these
    #       arguments should only really be used indirectly through the
    #       middleware decorator, where we can be more type safe
    @overload
    def __init__(
        self: wsgify[[], Request],
        func: _RequestHandler[Request, []] | None = None,
        RequestClass: None = None,
        args: tuple[()] = (),
        kwargs: None = None,
        middleware_wraps: None = None,
    ) -> None: ...
    @overload
    def __init__(
        self: wsgify[[], _RequestT_contra],  # pyright: ignore[reportInvalidTypeVarUse]  #11780
        func: _RequestHandler[_RequestT_contra, []] | None,
        RequestClass: type[_RequestT_contra],
        args: tuple[()] = (),
        kwargs: None = None,
        middleware_wraps: None = None,
    ) -> None: ...
    @overload
    def __init__(
        self: wsgify[[], _RequestT_contra],  # pyright: ignore[reportInvalidTypeVarUse]  #11780
        func: _RequestHandler[_RequestT_contra, []] | None = None,
        *,
        RequestClass: type[_RequestT_contra],
        args: tuple[()] = (),
        kwargs: None = None,
        middleware_wraps: None = None,
    ) -> None: ...
    @overload
    def __init__(
        self: wsgify[[_AppT_contra], Request],
        func: _Middleware[Request, _AppT_contra, []] | None = None,
        RequestClass: None = None,
        args: tuple[()] = (),
        kwargs: None = None,
        *,
        middleware_wraps: _AppT_contra,
    ) -> None: ...
    @overload
    def __init__(
        self: wsgify[[_AppT_contra], _RequestT_contra],  # pyright: ignore[reportInvalidTypeVarUse]  #11780
        func: _Middleware[_RequestT_contra, _AppT_contra, []] | None,
        RequestClass: type[_RequestT_contra],
        args: tuple[()] = (),
        kwargs: None = None,
        *,
        middleware_wraps: _AppT_contra,
    ) -> None: ...
    @overload
    def __init__(
        self: wsgify[[_AppT_contra], _RequestT_contra],  # pyright: ignore[reportInvalidTypeVarUse]  #11780
        func: _Middleware[_RequestT_contra, _AppT_contra, []] | None = None,
        *,
        RequestClass: type[_RequestT_contra],
        args: tuple[()] = (),
        kwargs: None = None,
        middleware_wraps: _AppT_contra,
    ) -> None: ...
    @overload
    def __get__(self, obj: None, type: type[_S]) -> _unbound_wsgify[_P, _S, _RequestT_contra]: ...
    @overload
    def __get__(self, obj: object, type: type | None = None) -> Self: ...
    @overload
    def __call__(self, env: WSGIEnvironment, /, start_response: StartResponse) -> Iterable[bytes]: ...
    @overload
    def __call__(self, func: _RequestHandler[_RequestT_contra, _P], /) -> Self: ...
    @overload
    def __call__(self, req: _RequestT_contra) -> _AnyResponse: ...
    @overload
    def __call__(self, req: _RequestT_contra, *args: _P.args, **kw: _P.kwargs) -> _AnyResponse: ...
    def get(self, url: str, **kw: Any) -> _AnyResponse: ...
    def post(
        self, url: str, POST: str | bytes | Mapping[Any, Any] | Mapping[Any, list[Any] | tuple[Any, ...]] | None = None, **kw: Any
    ) -> _AnyResponse: ...
    def request(self, url: str, **kw: Any) -> _AnyResponse: ...
    def call_func(self, req: _RequestT_contra, *args: _P.args, **kwargs: _P.kwargs) -> _AnyResponse: ...
    # technically this could bind different type vars, but we disallow it for safety
    def clone(self, func: _RequestHandler[_RequestT_contra, _P] | None = None, **kw: Never) -> Self: ...
    @property
    def undecorated(self) -> _RequestHandler[_RequestT_contra, _P] | None: ...
    @overload
    @classmethod
    def middleware(
        cls, middle_func: None = None, app: None | _AppT = None, *_: _P.args, **kw: _P.kwargs
    ) -> _UnboundMiddleware[_P, _AppT, Any]: ...
    @overload
    @classmethod
    def middleware(
        cls, middle_func: _MiddlewareCallable[_RequestT, _AppT, _P2], app: None = None
    ) -> _MiddlewareFactory[_P2, _AppT, _RequestT]: ...
    @overload
    @classmethod
    def middleware(
        cls, middle_func: _MiddlewareMethod[_RequestT, _AppT, _P2], app: None = None
    ) -> _MiddlewareFactory[_P2, _AppT, _RequestT]: ...
    @overload
    @classmethod
    def middleware(
        cls, middle_func: _MiddlewareMethod[_RequestT, _AppT, _P2], app: None = None, *_: _P2.args, **kw: _P2.kwargs
    ) -> _MiddlewareFactory[_P2, _AppT, _RequestT]: ...
    @overload
    @classmethod
    def middleware(
        cls, middle_func: _MiddlewareMethod[_RequestT, _AppT, _P2], app: _AppT
    ) -> type[wsgify[Concatenate[_AppT, _P2], _RequestT]]: ...
    @overload
    @classmethod
    def middleware(
        cls, middle_func: _MiddlewareMethod[_RequestT, _AppT, _P2], app: _AppT, *_: _P2.args, **kw: _P2.kwargs
    ) -> type[wsgify[Concatenate[_AppT, _P2], _RequestT]]: ...

@type_check_only
class _unbound_wsgify(wsgify[_P, _RequestT_contra], Generic[_P, _S, _RequestT_contra]):
    @overload  # type: ignore[override]
    def __call__(self, __self: _S, env: WSGIEnvironment, /, start_response: StartResponse) -> Iterable[bytes]: ...
    @overload
    def __call__(self, __self: _S, func: _RequestHandler[_RequestT_contra, _P], /) -> Self: ...
    @overload
    def __call__(self, __self: _S, /, req: _RequestT_contra) -> _AnyResponse: ...
    @overload
    def __call__(self, __self: _S, /, req: _RequestT_contra, *args: _P.args, **kw: _P.kwargs) -> _AnyResponse: ...

class _UnboundMiddleware(Generic[_P, _AppT_contra, _RequestT_contra]):
    wrapper_class: type[wsgify[Concatenate[_AppT_contra, _P], _RequestT_contra]]
    app: _AppT_contra | None
    kw: dict[str, Any]
    def __init__(
        self,
        wrapper_class: type[wsgify[Concatenate[_AppT_contra, _P], _RequestT_contra]],
        app: _AppT_contra | None,
        kw: dict[str, Any],
    ) -> None: ...
    @overload
    def __call__(self, func: None, app: _AppT_contra | None = None) -> Self: ...
    @overload
    def __call__(
        self, func: _Middleware[_RequestT_contra, _AppT_contra, _P], app: None = None
    ) -> wsgify[Concatenate[_AppT_contra, _P], _RequestT_contra]: ...
    @overload
    def __call__(
        self, func: _Middleware[_RequestT_contra, _AppT_contra, _P], app: _AppT_contra
    ) -> wsgify[Concatenate[_AppT_contra, _P], _RequestT_contra]: ...

class _MiddlewareFactory(Generic[_P, _AppT_contra, _RequestT_contra]):
    wrapper_class: type[wsgify[Concatenate[_AppT_contra, _P], _RequestT_contra]]
    middleware: _Middleware[_RequestT_contra, _AppT_contra, _P]
    kw: dict[str, Any]
    def __init__(
        self,
        wrapper_class: type[wsgify[Concatenate[_AppT_contra, _P], _RequestT_contra]],
        middleware: _Middleware[_RequestT_contra, _AppT_contra, _P],
        kw: dict[str, Any],
    ) -> None: ...
    # NOTE: Technically you are not allowed to pass args, but we give up all kinds
    #       of other safety if we don't use ParamSpec
    @overload
    def __call__(
        self, app: None = None, *_: _P.args, **config: _P.kwargs
    ) -> _MiddlewareFactory[[], _AppT_contra, _RequestT_contra]: ...
    @overload
    def __call__(self, app: _AppT_contra, *_: _P.args, **config: _P.kwargs) -> wsgify[[_AppT_contra], _RequestT_contra]: ...



############################################################################
# Copyright(c) Open Law Library. All rights reserved.                      #
# See ThirdPartyNotices.txt in the project root for additional notices.    #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License")           #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#     http: // www.apache.org/licenses/LICENSE-2.0                         #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
############################################################################
import asyncio
import logging
import re
from threading import Event
from typing import Any
from typing import Callable
from typing import List
from typing import Optional
from typing import Type
from typing import Union

from cattrs import Converter

from pygls.exceptions import PyglsError, JsonRpcException
from pygls.protocol import JsonRPCProtocol, default_converter


logger = logging.getLogger(__name__)


async def aio_readline(stop_event, reader, message_handler):
    CONTENT_LENGTH_PATTERN = re.compile(rb"^Content-Length: (\d+)\r\n$")

    # Initialize message buffer
    message = []
    content_length = 0

    while not stop_event.is_set():
        # Read a header line
        header = await reader.readline()
        if not header:
            break
        message.append(header)

        # Extract content length if possible
        if not content_length:
            match = CONTENT_LENGTH_PATTERN.fullmatch(header)
            if match:
                content_length = int(match.group(1))
                logger.debug("Content length: %s", content_length)

        # Check if all headers have been read (as indicated by an empty line \r\n)
        if content_length and not header.strip():
            # Read body
            body = await reader.readexactly(content_length)
            if not body:
                break
            message.append(body)

            # Pass message to protocol
            message_handler(b"".join(message))

            # Reset the buffer
            message = []
            content_length = 0


class JsonRPCClient:
    """Base JSON-RPC client."""

    def __init__(
        self,
        protocol_cls: Type[JsonRPCProtocol] = JsonRPCProtocol,
        converter_factory: Callable[[], Converter] = default_converter,
    ):
        # Strictly speaking `JsonRPCProtocol` wants a `LanguageServer`, not a
        # `JsonRPCClient`. However there similar enough for our purposes, which is
        # that this client will mostly be used in testing contexts.
        self.protocol = protocol_cls(self, converter_factory())  # type: ignore

        self._server: Optional[asyncio.subprocess.Process] = None
        self._stop_event = Event()
        self._async_tasks: List[asyncio.Task] = []

    @property
    def stopped(self) -> bool:
        """Return ``True`` if the client has been stopped."""
        return self._stop_event.is_set()

    def feature(
        self,
        feature_name: str,
        options: Optional[Any] = None,
    ):
        """Decorator used to register LSP features.

        Example
        -------
        ::

           import logging
           from pygls.client import JsonRPCClient

           ls = JsonRPCClient()

           @ls.feature('window/logMessage')
           def completions(ls, params):
               logging.info("%s", params.message)
        """
        return self.protocol.fm.feature(feature_name, options)

    async def start_io(self, cmd: str, *args, **kwargs):
        """Start the given server and communicate with it over stdio."""

        logger.debug("Starting server process: %s", " ".join([cmd, *args]))
        server = await asyncio.create_subprocess_exec(
            cmd,
            *args,
            stdout=asyncio.subprocess.PIPE,
            stdin=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            **kwargs,
        )

        self.protocol.connection_made(server.stdin)  # type: ignore
        connection = asyncio.create_task(
            aio_readline(self._stop_event, server.stdout, self.protocol.data_received)
        )
        notify_exit = asyncio.create_task(self._server_exit())

        self._server = server
        self._async_tasks.extend([connection, notify_exit])

    async def _server_exit(self):
        if self._server is not None:
            await self._server.wait()
            logger.debug(
                "Server process %s exited with return code: %s",
                self._server.pid,
                self._server.returncode,
            )
            await self.server_exit(self._server)
            self._stop_event.set()

    async def server_exit(self, server: asyncio.subprocess.Process):
        """Called when the server process exits."""

    def _report_server_error(
        self, error: Exception, source: Union[PyglsError, JsonRpcException]
    ):
        try:
            self.report_server_error(error, source)
        except Exception:
            logger.error("Unable to report error", exc_info=True)

    def report_server_error(
        self, error: Exception, source: Union[PyglsError, JsonRpcException]
    ):
        """Called when the server does something unexpected e.g. respond with malformed
        JSON."""

    async def stop(self):
        self._stop_event.set()

        if self._server is not None and self._server.returncode is None:
            logger.debug("Terminating server process: %s", self._server.pid)
            self._server.terminate()

        if len(self._async_tasks) > 0:
            await asyncio.gather(*self._async_tasks)



from _typeshed import Incomplete
from collections.abc import Generator

WHEEL_NAME: Incomplete
NAMESPACE_PACKAGE_INIT: str

def unpack(src_dir, dst_dir) -> None: ...
def disable_info_traces() -> Generator[None]: ...

class Wheel:
    filename: Incomplete
    def __init__(self, filename) -> None: ...
    def tags(self): ...
    def is_compatible(self): ...
    def egg_name(self): ...
    def get_dist_info(self, zf): ...
    def install_as_egg(self, destination_eggdir) -> None: ...



from collections.abc import Collection

from authlib.jose.rfc7517 import Key

class KeySet:
    keys: Collection[Key]
    def __init__(self, keys) -> None: ...
    def as_dict(self, is_private: bool = False, **params): ...
    def as_json(self, is_private: bool = False, **params): ...
    def find_by_kid(self, kid, **params): ...



from _typeshed import Unused
from collections.abc import Iterable
from typing import Any

from asgiref.typing import HTTPDisconnectEvent, HTTPRequestEvent, HTTPScope
from channels.consumer import AsyncConsumer

class AsyncHttpConsumer(AsyncConsumer):
    body: list[bytes]
    scope: HTTPScope  # type: ignore[assignment]

    def __init__(self, *args: Unused, **kwargs: Unused) -> None: ...
    async def send_headers(self, *, status: int = 200, headers: Iterable[tuple[bytes, bytes]] | None = None) -> None: ...
    async def send_body(self, body: bytes, *, more_body: bool = False) -> None: ...
    async def send_response(self, status: int, body: bytes, **kwargs: Any) -> None: ...
    async def handle(self, body: bytes) -> None: ...
    async def disconnect(self) -> None: ...
    async def http_request(self, message: HTTPRequestEvent) -> None: ...
    async def http_disconnect(self, message: HTTPDisconnectEvent) -> None: ...



from typing import Final, Literal

WEEKDAYS: Final = ("Ravivāra", "Somavāra", "Maṅgalavāra", "Budhavāra", "Guruvāra", "Śukravāra", "Śanivāra")
MONTHS: Final[tuple[str, ...]]
HAVE_31_DAYS: Final = (2, 3, 4, 5, 6)
HAVE_30_DAYS: Final = (7, 8, 9, 10, 11, 12)
SAKA_EPOCH: Final = 78

def to_jd(year: int, month: int, day: int) -> float: ...
def from_jd(jd: float) -> tuple[int, int, int]: ...
def from_gregorian(year: int, month: int, day: int) -> tuple[int, int, int]: ...
def to_gregorian(year: int, month: int, day: int) -> tuple[int, int, int]: ...
def month_length(year: int, month: int) -> Literal[30, 31]: ...
def monthcalendar(year: int, month: int) -> list[list[int | None]]: ...
def format(year: int, month: int, day: int) -> str: ...



from collections.abc import Mapping, Sequence
from xml.etree.ElementTree import Element

from markdown import inlinepatterns, util
from markdown.core import Markdown
from markdown.extensions import Extension
from markdown.inlinepatterns import HtmlInlineProcessor

punctClass: str
endOfWordClass: str
closeClass: str
openingQuotesBase: str
substitutions: Mapping[str, str]
singleQuoteStartRe: str
doubleQuoteStartRe: str
doubleQuoteSetsRe: str
singleQuoteSetsRe: str
doubleQuoteSetsRe2: str
singleQuoteSetsRe2: str
decadeAbbrRe: str
openingDoubleQuotesRegex: str
closingDoubleQuotesRegex: str
closingDoubleQuotesRegex2: str
openingSingleQuotesRegex: str
closingSingleQuotesRegex: str
closingSingleQuotesRegex2: str
remainingSingleQuotesRegex: str
remainingDoubleQuotesRegex: str
HTML_STRICT_RE: str

class SubstituteTextPattern(HtmlInlineProcessor):
    replace: Sequence[int | str | Element]
    def __init__(self, pattern: str, replace: Sequence[int | str | Element], md: Markdown) -> None: ...

class SmartyExtension(Extension):
    substitutions: dict[str, str]
    def __init__(self, **kwargs) -> None: ...
    def educateDashes(self, md: Markdown) -> None: ...
    def educateEllipses(self, md: Markdown) -> None: ...
    def educateAngledQuotes(self, md: Markdown) -> None: ...
    def educateQuotes(self, md: Markdown) -> None: ...
    inlinePatterns: util.Registry[inlinepatterns.Pattern]

def makeExtension(**kwargs) -> SmartyExtension: ...



from _typeshed import Incomplete
from collections.abc import Callable
from re import Pattern
from typing import Any, Final, Generic, TypedDict, TypeVar, overload, type_check_only
from typing_extensions import TypeAlias

from ..rfc7517 import KeySet
from .claims import JWTClaims

_T = TypeVar("_T")

_LoadKey: TypeAlias = Callable[[Incomplete, Incomplete], Incomplete]

class JsonWebToken:
    SENSITIVE_NAMES: Final[tuple[str, ...]]
    SENSITIVE_VALUES: Final[Pattern[str]]

    def __init__(self, algorithms, private_headers=None) -> None: ...
    def check_sensitive_data(self, payload) -> None: ...
    def encode(self, header, payload, key, check: bool = True): ...
    @overload
    def decode(
        self,
        s: str | bytes,
        key: _LoadKey | KeySet | tuple[Incomplete, ...] | list[Incomplete] | str,
        claims_cls: None = None,
        claims_options=None,
        claims_params=None,
    ) -> JWTClaims: ...
    @overload
    def decode(
        self,
        s: str | bytes,
        key: _LoadKey | KeySet | tuple[Incomplete, ...] | list[Incomplete] | str,
        claims_cls: type[_T],
        claims_options=None,
        claims_params=None,
    ) -> _T: ...

def decode_payload(bytes_payload): ...

_TL = TypeVar("_TL", bound=tuple[Any, ...] | list[Any])

@type_check_only
class _Keys(TypedDict, Generic[_TL]):
    keys: _TL

@overload
def prepare_raw_key(raw: KeySet) -> KeySet: ...
@overload
def prepare_raw_key(raw: str) -> dict[str, Any] | str: ...  # dict is a JSON object
@overload
def prepare_raw_key(raw: _TL) -> _Keys[_TL]: ...
def find_encode_key(key, header): ...
def create_load_key(key: KeySet | _Keys[Incomplete] | Incomplete) -> _LoadKey: ...



def bind_operation(
    version, authentication, name: str = "", password=None, sasl_mechanism=None, sasl_credentials=None, auto_encode: bool = False
): ...
def bind_request_to_dict(request): ...
def bind_response_operation(
    result_code, matched_dn: str = "", diagnostic_message: str = "", referral=None, server_sasl_credentials=None
): ...
def bind_response_to_dict(response): ...
def sicily_bind_response_to_dict(response): ...
def bind_response_to_dict_fast(response): ...
def sicily_bind_response_to_dict_fast(response): ...



from collections.abc import Iterable
from datetime import timedelta
from logging import Logger
from re import Match, Pattern
from typing import Any, Final, Literal, TypedDict, TypeVar, overload, type_check_only
from typing_extensions import TypeAlias

import flask

_IterableT = TypeVar("_IterableT", bound=Iterable[Any])
_T = TypeVar("_T")
_MultiDict: TypeAlias = Any  # werkzeug is not part of typeshed

@type_check_only
class _Options(TypedDict, total=False):
    resources: dict[str, dict[str, Any]] | list[str] | str | None
    origins: str | list[str] | None
    methods: str | list[str] | None
    expose_headers: str | list[str] | None
    allow_headers: str | list[str] | None
    supports_credentials: bool | None
    max_age: timedelta | int | str | None
    send_wildcard: bool | None
    vary_header: bool | None
    automatic_options: bool | None
    intercept_exceptions: bool | None
    always_send: bool | None

LOG: Logger

ACL_ORIGIN: Final = "Access-Control-Allow-Origin"
ACL_METHODS: Final = "Access-Control-Allow-Methods"
ACL_ALLOW_HEADERS: Final = "Access-Control-Allow-Headers"
ACL_EXPOSE_HEADERS: Final = "Access-Control-Expose-Headers"
ACL_CREDENTIALS: Final = "Access-Control-Allow-Credentials"
ACL_MAX_AGE: Final = "Access-Control-Max-Age"
ACL_RESPONSE_PRIVATE_NETWORK: Final = "Access-Control-Allow-Private-Network"
ACL_REQUEST_METHOD: Final = "Access-Control-Request-Method"
ACL_REQUEST_HEADERS: Final = "Access-Control-Request-Headers"
ACL_REQUEST_HEADER_PRIVATE_NETWORK: Final = "Access-Control-Request-Private-Network"
ALL_METHODS: Final[list[str]]
CONFIG_OPTIONS: Final[list[str]]
FLASK_CORS_EVALUATED: Final = "_FLASK_CORS_EVALUATED"
RegexObject: Final[type[Pattern[str]]]
DEFAULT_OPTIONS: Final[_Options]

def parse_resources(resources: dict[str, _Options] | Iterable[str] | str | Pattern[str]) -> list[tuple[str, _Options]]: ...
def get_regexp_pattern(regexp: str | Pattern[str]) -> str: ...
def get_cors_origins(options: _Options, request_origin: str | None) -> list[str] | None: ...
def get_allow_headers(options: _Options, acl_request_headers: str | None) -> str | None: ...
def get_cors_headers(options: _Options, request_headers: dict[str, Any], request_method: str) -> _MultiDict: ...
def set_cors_headers(resp: flask.Response, options: _Options) -> flask.Response: ...
@overload
def probably_regex(maybe_regex: Pattern[str]) -> Literal[True]: ...
@overload
def probably_regex(maybe_regex: str) -> bool: ...
def re_fix(reg: str) -> str: ...
def try_match_any_pattern(inst: str, patterns: Iterable[str | Pattern[str]], caseSensitive: bool = True) -> bool: ...
def try_match_pattern(value: str, pattern: str | Pattern[str], caseSensitive: bool = True) -> bool | Match[str]: ...
def get_cors_options(appInstance: flask.Flask | None, *dicts: _Options) -> _Options: ...
def get_app_kwarg_dict(appInstance: flask.Flask | None = None) -> _Options: ...
def flexible_str(obj: object) -> str | None: ...
def serialize_option(options_dict: _Options, key: str, upper: bool = False) -> None: ...
@overload
def ensure_iterable(inst: str) -> list[str]: ...  # type: ignore[overload-overlap]
@overload
def ensure_iterable(inst: _IterableT) -> _IterableT: ...  # type: ignore[overload-overlap]
@overload
def ensure_iterable(inst: _T) -> list[_T]: ...
def sanitize_regex_param(param: str | list[str]) -> list[str]: ...
def serialize_options(opts: _Options) -> _Options: ...



from __future__ import annotations

from typing import Any


# The following should pass without error (see #6661):
class Diagnostic:
    def __reduce__(self) -> str | tuple[Any, ...]:
        res = super().__reduce__()
        if isinstance(res, tuple) and len(res) >= 3:
            res[2]["_info"] = 42

        return res



from networkx.classes.graph import Graph, _Node
from networkx.utils.backends import _dispatchable

__all__ = ["immediate_dominators", "dominance_frontiers"]

@_dispatchable
def immediate_dominators(G: Graph[_Node], start: _Node): ...
@_dispatchable
def dominance_frontiers(G: Graph[_Node], start: _Node): ...



from _typeshed import StrPath
from collections.abc import Collection

from xdg.IniFile import IniFile

class IconTheme(IniFile):
    def __init__(self) -> None: ...
    dir: str
    name: str
    comment: str
    directories: list[str]
    type: str
    def parse(self, file: StrPath) -> None: ...  # type: ignore[override]
    def getDir(self) -> str: ...
    def getName(self) -> str: ...
    def getComment(self) -> str: ...
    def getInherits(self) -> list[str]: ...
    def getDirectories(self) -> list[str]: ...
    def getScaledDirectories(self) -> list[str]: ...
    def getHidden(self) -> bool: ...
    def getExample(self) -> str: ...
    def getSize(self, directory: StrPath) -> int: ...
    def getContext(self, directory: StrPath) -> str: ...
    def getType(self, directory: StrPath) -> str: ...
    def getMaxSize(self, directory: StrPath) -> int: ...
    def getMinSize(self, directory: StrPath) -> int: ...
    def getThreshold(self, directory: StrPath) -> int: ...
    def getScale(self, directory: StrPath) -> int: ...
    def checkExtras(self) -> None: ...
    def checkGroup(self, group: str) -> None: ...
    def checkKey(self, key: str, value: str, group: str) -> None: ...

class IconData(IniFile):
    def __init__(self) -> None: ...
    def parse(self, file: StrPath) -> None: ...  # type: ignore[override]
    def getDisplayName(self) -> str: ...
    def getEmbeddedTextRectangle(self) -> list[int]: ...
    def getAttachPoints(self) -> list[tuple[int, int]]: ...
    def checkExtras(self) -> None: ...
    def checkGroup(self, group: str) -> None: ...
    def checkKey(self, key: str, value: str, group: str) -> None: ...

icondirs: list[str]
themes: list[IconTheme]
theme_cache: dict[str, IconTheme]
dir_cache: dict[str, tuple[str, float, float]]
icon_cache: dict[tuple[str, int, str, tuple[str, ...]], tuple[float, str]]

def getIconPath(
    iconname: str, size: int | None = None, theme: str | None = None, extensions: Collection[str] = ["png", "svg", "xpm"]
) -> str: ...
def getIconData(path: str) -> IconData: ...
def LookupIcon(iconname: str, size: int, theme: str, extensions: Collection[str]) -> str: ...
def DirectoryMatchesSize(subdir: str, iconsize: int, theme: str) -> bool: ...
def DirectorySizeDistance(subdir: str, iconsize: int, theme: str) -> int: ...



from _typeshed import Incomplete

from antlr4.atn.LexerAction import LexerAction as LexerAction, LexerIndexedCustomAction as LexerIndexedCustomAction
from antlr4.InputStream import InputStream as InputStream

class LexerActionExecutor:
    __slots__ = ("lexerActions", "hashCode")
    lexerActions: Incomplete
    hashCode: Incomplete
    def __init__(self, lexerActions: list[LexerAction] = []) -> None: ...
    @staticmethod
    def append(lexerActionExecutor: LexerActionExecutor, lexerAction: LexerAction): ...
    def fixOffsetBeforeMatch(self, offset: int): ...
    def execute(self, lexer, input: InputStream, startIndex: int): ...
    def __hash__(self): ...
    def __eq__(self, other): ...



from _typeshed import Incomplete
from collections.abc import Generator
from typing import Final, Literal

from networkx.classes.graph import Graph, _Node
from networkx.utils.backends import _dispatchable

__all__ = [
    "write_graphml",
    "read_graphml",
    "generate_graphml",
    "write_graphml_xml",
    "write_graphml_lxml",
    "parse_graphml",
    "GraphMLWriter",
    "GraphMLReader",
]

def write_graphml_xml(
    G: Graph[_Node],
    path,
    encoding: str = "utf-8",
    prettyprint: bool = True,
    infer_numeric_types: bool = False,
    named_key_ids: bool = False,
    edge_id_from_attribute=None,
) -> None: ...
def write_graphml_lxml(
    G: Graph[_Node],
    path,
    encoding: str = "utf-8",
    prettyprint: bool = True,
    infer_numeric_types: bool = False,
    named_key_ids: bool = False,
    edge_id_from_attribute=None,
): ...
def generate_graphml(
    G: Graph[_Node], encoding: str = "utf-8", prettyprint: bool = True, named_key_ids: bool = False, edge_id_from_attribute=None
) -> Generator[Incomplete, Incomplete, None]: ...
@_dispatchable
def read_graphml(path, node_type=..., edge_key_type=..., force_multigraph: bool = False): ...
@_dispatchable
def parse_graphml(graphml_string, node_type=..., edge_key_type=..., force_multigraph: bool = False): ...

class GraphML:
    NS_GRAPHML: Final[str]
    NS_XSI: Final[str]
    NS_Y: Final[str]
    SCHEMALOCATION: Final[str]
    xml_type: Incomplete
    python_type: Incomplete
    def construct_types(self) -> None: ...
    convert_bool: Final[dict[Literal["true", "false", "0", 0, "1", 1], bool]]
    def get_xml_type(self, key): ...

class GraphMLWriter(GraphML):
    myElement: Incomplete
    infer_numeric_types: Incomplete
    prettyprint: Incomplete
    named_key_ids: Incomplete
    edge_id_from_attribute: Incomplete
    encoding: Incomplete
    xml: Incomplete
    keys: Incomplete
    attributes: Incomplete
    attribute_types: Incomplete
    def __init__(
        self,
        graph=None,
        encoding: str = "utf-8",
        prettyprint: bool = True,
        infer_numeric_types: bool = False,
        named_key_ids: bool = False,
        edge_id_from_attribute=None,
    ) -> None: ...
    def attr_type(self, name, scope, value): ...
    def get_key(self, name, attr_type, scope, default): ...
    def add_data(self, name, element_type, value, scope: str = "all", default=None): ...
    def add_attributes(self, scope, xml_obj, data, default) -> None: ...
    def add_nodes(self, G: Graph[_Node], graph_element) -> None: ...
    def add_edges(self, G: Graph[_Node], graph_element) -> None: ...
    def add_graph_element(self, G: Graph[_Node]) -> None: ...
    def add_graphs(self, graph_list) -> None: ...
    def dump(self, stream) -> None: ...
    def indent(self, elem, level: int = 0) -> None: ...

class IncrementalElement:
    xml: Incomplete
    prettyprint: Incomplete
    def __init__(self, xml, prettyprint) -> None: ...
    def append(self, element) -> None: ...

class GraphMLWriterLxml(GraphMLWriter):
    myElement: Incomplete
    named_key_ids: Incomplete
    edge_id_from_attribute: Incomplete
    infer_numeric_types: Incomplete
    xml: Incomplete
    keys: Incomplete
    attribute_types: Incomplete
    def __init__(
        self,
        path,
        graph=None,
        encoding: str = "utf-8",
        prettyprint: bool = True,
        infer_numeric_types: bool = False,
        named_key_ids: bool = False,
        edge_id_from_attribute=None,
    ) -> None: ...
    def add_graph_element(self, G: Graph[_Node]) -> None: ...
    def add_attributes(self, scope, xml_obj, data, default) -> None: ...
    def dump(self, stream=None) -> None: ...

write_graphml = write_graphml_lxml

class GraphMLReader(GraphML):
    node_type: Incomplete
    edge_key_type: Incomplete
    multigraph: Incomplete
    edge_ids: Incomplete
    def __init__(self, node_type=..., edge_key_type=..., force_multigraph: bool = False) -> None: ...
    xml: Incomplete
    def __call__(self, path=None, string=None) -> Generator[Incomplete, None, None]: ...
    def make_graph(self, graph_xml, graphml_keys, defaults, G=None): ...
    def add_node(self, G: Graph[_Node], node_xml, graphml_keys, defaults) -> None: ...
    def add_edge(self, G: Graph[_Node], edge_element, graphml_keys) -> None: ...
    def decode_data_elements(self, graphml_keys, obj_xml): ...
    def find_graphml_keys(self, graph_element): ...



from collections.abc import Callable
from typing import TypeVar
from typing_extensions import TypeAlias

from Xlib.error import XError
from Xlib.protocol.rq import Request

_T = TypeVar("_T")
ErrorHandler: TypeAlias = Callable[[XError, Request | None], _T]



from _typeshed import Incomplete
from collections.abc import Generator, Iterable
from os import PathLike
from typing import TypeVar
from typing_extensions import TypeAlias

_DirNode: TypeAlias = Incomplete  # scantree.DirNode
_RecursionPath: TypeAlias = Incomplete  # scantree.RecursionPath
_RP = TypeVar("_RP", bound=_RecursionPath)

__all__ = [
    "__version__",
    "algorithms_guaranteed",
    "algorithms_available",
    "dirhash",
    "dirhash_impl",
    "included_paths",
    "Filter",
    "get_match_patterns",
    "Protocol",
]

__version__: str
algorithms_guaranteed: set[str]
algorithms_available: set[str]

def dirhash(
    directory: str | PathLike[str],
    algorithm: str,
    match: Iterable[str] = ("*",),
    ignore: Iterable[str] | None = None,
    linked_dirs: bool = True,
    linked_files: bool = True,
    empty_dirs: bool = False,
    entry_properties: Iterable[str] = ("name", "data"),
    allow_cyclic_links: bool = False,
    chunk_size: int = 1048576,
    jobs: int = 1,
) -> str: ...
def dirhash_impl(
    directory: str | PathLike[str],
    algorithm: str,
    filter_: Filter | None = None,
    protocol: Protocol | None = None,
    chunk_size: int = 1048576,
    jobs: int = 1,
) -> str: ...
def included_paths(
    directory: str | PathLike[str],
    match: Iterable[str] = ("*",),
    ignore: Iterable[str] | None = None,
    linked_dirs: bool = True,
    linked_files: bool = True,
    empty_dirs: bool = False,
    allow_cyclic_links: bool = False,
) -> list[str]: ...

class Filter:
    linked_dirs: bool
    linked_files: bool
    empty_dirs: bool

    def __init__(
        self,
        match_patterns: Iterable[str] | None = None,
        linked_dirs: bool = True,
        linked_files: bool = True,
        empty_dirs: bool = False,
    ) -> None: ...
    @property
    def match_patterns(self) -> tuple[str, ...]: ...
    def include(self, recursion_path: _RecursionPath) -> bool: ...
    def match_file(self, filepath: str | PathLike[str]) -> bool: ...
    def __call__(self, paths: Iterable[_RP]) -> Generator[_RP, None, None]: ...

def get_match_patterns(
    match: Iterable[str] | None = None,
    ignore: Iterable[str] | None = None,
    ignore_extensions: Iterable[str] | None = None,
    ignore_hidden: bool = False,
) -> list[str]: ...

class Protocol:
    class EntryProperties:
        NAME: str
        DATA: str
        IS_LINK: str
        options: set[str]

    entry_properties: Iterable[str]
    allow_cyclic_links: bool
    def __init__(self, entry_properties: Iterable[str] = ("name", "data"), allow_cyclic_links: bool = False) -> None: ...
    def get_descriptor(self, dir_node: _DirNode) -> str: ...



from _typeshed import Incomplete, Unused
from collections.abc import Callable, Generator
from typing import Final

from networkx.classes.graph import Graph, _Node
from networkx.utils.backends import _dispatchable

__all__ = [
    "greedy_color",
    "strategy_connected_sequential",
    "strategy_connected_sequential_bfs",
    "strategy_connected_sequential_dfs",
    "strategy_independent_set",
    "strategy_largest_first",
    "strategy_random_sequential",
    "strategy_saturation_largest_first",
    "strategy_smallest_last",
]

@_dispatchable
def strategy_largest_first(G: Graph[_Node], colors: Unused): ...
@_dispatchable
def strategy_random_sequential(G: Graph[_Node], colors: Unused, seed=None): ...
@_dispatchable
def strategy_smallest_last(G: Graph[_Node], colors: Unused): ...
@_dispatchable
def strategy_independent_set(G: Graph[_Node], colors: Unused) -> Generator[Incomplete, Incomplete, None]: ...
@_dispatchable
def strategy_connected_sequential_bfs(G: Graph[_Node], colors): ...
@_dispatchable
def strategy_connected_sequential_dfs(G: Graph[_Node], colors): ...
@_dispatchable
def strategy_connected_sequential(
    G: Graph[_Node], colors: Unused, traversal: str = "bfs"
) -> Generator[Incomplete, None, None]: ...
@_dispatchable
def strategy_saturation_largest_first(G: Graph[_Node], colors) -> Generator[Incomplete, None, Incomplete]: ...

STRATEGIES: Final[dict[str, Callable[..., Incomplete]]]

@_dispatchable
def greedy_color(G: Graph[_Node], strategy="largest_first", interchange: bool = False): ...



from _typeshed import Incomplete
from collections.abc import Iterable

from networkx.classes.graph import Graph, _Node
from networkx.exception import NetworkXException
from networkx.utils.backends import _dispatchable

__all__ = ["from_nested_tuple", "from_prufer_sequence", "NotATree", "to_nested_tuple", "to_prufer_sequence"]

class NotATree(NetworkXException): ...

@_dispatchable
def to_nested_tuple(T: Graph[_Node], root: _Node, canonical_form: bool = False): ...
@_dispatchable
def from_nested_tuple(sequence: tuple[Incomplete, ...], sensible_relabeling: bool = False): ...
@_dispatchable
def to_prufer_sequence(T: Graph[_Node]) -> list[Incomplete]: ...
@_dispatchable
def from_prufer_sequence(sequence: Iterable[Incomplete]): ...
